# Copyright 2022-2024 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging
import math

import zuul.provider.schema as provider_schema
from zuul.lib.voluputil import (
    AsList,
    Nullable,
    Optional,
    Required,
    assemble,
    discriminate,
)

import voluptuous as vs

from zuul.driver.openstack.openstackendpoint import (
    OpenstackCreateStateMachine,
    OpenstackDeleteStateMachine,
)
from zuul.model import QuotaInformation
from zuul.provider import (
    BaseProvider,
    BaseProviderFlavor,
    BaseProviderImage,
    BaseProviderLabel,
    BaseProviderSchema,
)


class OpenstackProviderImage(BaseProviderImage):
    # This is used here and in flavors and labels
    inheritable_openstack_image_schema = assemble(
        vs.Schema({
            Optional(
                'volume-size',
                doc="""\
                When booting an image from volume, this indicates the
                size of the created volume, in GB.
                """,
            ): Nullable(int),
        }),
        provider_schema.cloud_image,
    )
    openstack_cloud_schema = vs.Schema({
        Required(
            'image-id',
            doc="""\
            The ID of the cloud provider's image.
            """,
        ): str,
        Optional(
            'config-drive',
            doc="""\
            Whether config drive should be used for the cloud
            image.
            """,
            default=True): bool,
    })
    cloud_schema = vs.All(
        assemble(
            BaseProviderImage.cloud_schema,
            openstack_cloud_schema,
            inheritable_openstack_image_schema,
        ),
    )
    inheritable_openstack_zuul_schema = vs.Schema({
        Optional(
            'config-drive',
            doc="""\
            Whether config drive should be used for the cloud
            image.
            """,
            default=True): bool,
    })
    zuul_schema = assemble(
        BaseProviderImage.zuul_schema,
        inheritable_openstack_image_schema,
        inheritable_openstack_zuul_schema,
    )
    inheritable_cloud_schema = assemble(
        BaseProviderImage.inheritable_cloud_schema,
        inheritable_openstack_image_schema,
    )
    inheritable_zuul_schema = assemble(
        BaseProviderImage.inheritable_zuul_schema,
        inheritable_openstack_image_schema,
        inheritable_openstack_zuul_schema,
    )
    schema = vs.Union(
        cloud_schema, zuul_schema,
        discriminant=discriminate(
            lambda val, alt: val['type'] == alt['type'].validators[0])
    )
    inheritable_schema = assemble(
        inheritable_cloud_schema,
        inheritable_zuul_schema,
    )

    def __init__(self, image_config, provider_config, image_format):
        self.image_id = None
        super().__init__(image_config, provider_config)
        self.format = image_format
        # Implement provider defaults
        if self.connection_type is None:
            self.connection_type = 'ssh'
        if self.connection_port is None:
            self.connection_port = 22


class OpenstackProviderFlavor(BaseProviderFlavor):
    openstack_flavor_schema = vs.Schema({
        Required(
            'flavor-name',
            doc="""\
            Name or id of the OpenStack flavor to use.
            """,
        ): str,
    })

    inheritable_schema = assemble(
        BaseProviderFlavor.inheritable_schema,
        OpenstackProviderImage.inheritable_openstack_image_schema,
        provider_schema.cloud_flavor,
    )
    schema = assemble(
        BaseProviderFlavor.schema,
        provider_schema.cloud_flavor,
        OpenstackProviderImage.inheritable_openstack_image_schema,
        openstack_flavor_schema,
    )


class OpenstackProviderLabel(BaseProviderLabel):
    inheritable_openstack_label_schema = vs.Schema({
        Optional(
            'az',
            doc="""\
            Servers will be assigned to the specified availibility
            zone.  If omitted, one will be chosen at random.
            """,
        ): Nullable(str),
        Optional(
            'auto-floating-ip',
            doc="""\
            Whether to automatically allocate and assign a floating IP
            for each node.
            """,
            default=True): bool,
        Optional(
            'boot-from-volume',
            doc="""\
            Whether to create a volume from the image and boot the
            node from it.
            """,
            default=False): bool,
        Optional(
            'networks', default=[],
            doc="""\
            The OpenStack networks to associate with the node.
            """,
        ): AsList(str),
        Optional(
            'security-groups',
            doc="""\
            Specify custom networks to be attached to each
            node.  Specify the name or id of the network as a string.
            """,
            default=[]): AsList(str),
    })
    inheritable_schema = assemble(
        BaseProviderLabel.inheritable_schema,
        OpenstackProviderImage.inheritable_openstack_image_schema,
        provider_schema.ssh_label,
        inheritable_openstack_label_schema,
    )
    schema = assemble(
        BaseProviderLabel.schema,
        OpenstackProviderImage.inheritable_openstack_image_schema,
        provider_schema.ssh_label,
        inheritable_openstack_label_schema,
    )


class OpenstackProviderSchema(BaseProviderSchema):
    def getLabelSchema(self):
        return OpenstackProviderLabel.schema

    def getImageSchema(self):
        return OpenstackProviderImage.schema

    def getFlavorSchema(self):
        return OpenstackProviderFlavor.schema

    def getInheritableLabelSchema(self):
        return OpenstackProviderLabel.inheritable_schema

    def getInheritableImageSchema(self):
        return OpenstackProviderImage.inheritable_schema

    def getInheritableZuulImageSchema(self):
        return OpenstackProviderImage.inheritable_zuul_schema

    def getInheritableCloudImageSchema(self):
        return OpenstackProviderImage.inheritable_cloud_schema

    def getInheritableFlavorSchema(self):
        return OpenstackProviderFlavor.inheritable_schema

    def getProviderSchema(self):
        schema = super().getProviderSchema()

        resource_limits = {
            Optional(
                'instances',
                default=vs.UNDEFINED,
                doc="""The number of instances.""",
            ): int,
            Optional(
                'cores',
                default=vs.UNDEFINED,
                doc="""The number of cores.""",
            ): int,
            Optional(
                'ram',
                default=vs.UNDEFINED,
                doc="""The amount of ram, in MiB.""",
            ): int,
            Optional(
                'volumes',
                default=vs.UNDEFINED,
                doc="""The number of volumes.""",
            ): int,
            Optional(
                'volume-gb',
                default=vs.UNDEFINED,
                doc="""The amount of volume storage in GB.""",
            ): int,
        }

        openstack_provider_schema = vs.Schema({
            Optional(
                'region',
                doc="""\
                The region name if the provider cloud has multiple
                regions.
                """,
            ): Nullable(str),
            Optional(
                'resource-limits',
                doc="""\
                Resource limits for this provider.  Configure these
                values to cause Zuul to attempt to limit the resource
                usage.  This can be used to limit Zuul's usage to a
                level below the cloud quota.
                """,
                default=dict()): resource_limits,
            Optional(
                'floating-ip-cleanup',
                doc="""\
                If set to ``true``, Zuul will behave as if it is the
                only user of the OpenStack project and will attempt to
                clean unattached floating IPs that may have leaked.
                """,
                default=False): bool,
            Optional(
                'port-cleanup-interval',
                doc="""\
                If greater than 0, Zuul will behave as if it is the
                only user of the OpenStack project and will attempt to
                clean ports in ``DOWN`` state after the cleanup
                interval has elapsed.  This value may be reduced if
                the instance spawn time on the provider is reliably
                quicker.
                """,
                default=600): int,
        })

        return assemble(
            schema,
            openstack_provider_schema,
            doc="""\
            The attributes available for configuring an OpenStack provider
            are below.
            """,
        )


class OpenstackProvider(BaseProvider, subclass_id='openstack'):
    log = logging.getLogger("zuul.OpenstackProvider")
    schema = OpenstackProviderSchema().getProviderSchema()

    @property
    def endpoint(self):
        ep = getattr(self, '_endpoint', None)
        if ep:
            return ep
        self._set(_endpoint=self.getEndpoint())
        return self._endpoint

    def parseImage(self, image_config, provider_config, connection):
        # We are not fully constructed yet at this point, so we need
        # to peek to get the region and endpoint.
        region = provider_config.get('region')
        endpoint = connection.driver._getEndpoint(
            self.zk_client, connection, region, self.system_id)
        return OpenstackProviderImage(
            image_config, provider_config,
            image_format=endpoint.getImageFormat())

    def parseFlavor(self, flavor_config, provider_config, connection):
        return OpenstackProviderFlavor(flavor_config, provider_config)

    def parseLabel(self, label_config, provider_config, connection):
        return OpenstackProviderLabel(label_config, provider_config)

    def getEndpoint(self):
        return self.driver.getEndpoint(self)

    def getCreateStateMachine(self, node, image_external_id, log):
        # TODO: decide on a method of producing a hostname
        # that is max 15 chars.
        hostname = f"np{node.uuid[:13]}"
        label = self.labels[node.label]
        flavor = self.flavors[label.flavor]
        image = self.images[label.image]
        return OpenstackCreateStateMachine(
            self.endpoint,
            node,
            hostname,
            label,
            flavor,
            image,
            image_external_id,
            node.tags,
            log)

    def getDeleteStateMachine(self, node, log):
        return OpenstackDeleteStateMachine(self.endpoint, node, log)

    def getEndpointLimits(self):
        limits = self.endpoint.quota_cache.getLimits()
        if limits is None:
            limits = {}
        else:
            limits = limits.quota
        return QuotaInformation(default=math.inf, **limits)

    def getQuotaForLabel(self, label):
        flavor = self.flavors[label.flavor]
        return self.endpoint.getQuotaForLabel(label, flavor)

    def refreshQuotaForLabel(self, label, update):
        flavor = self.flavors[label.flavor]
        return self.endpoint.refreshQuotaForLabel(label, flavor, update)

    def getImageUploadJob(self, provider_image, image_name,
                          image_format, metadata, md5, sha256):
        return self.endpoint.getImageUploadJob(
            provider_image, image_name,
            image_format, metadata, md5, sha256)

    def deleteImage(self, external_id):
        self.endpoint.deleteImage(external_id)



# Copyright 2023 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import datetime

from apscheduler.triggers.cron import CronTrigger


class ZuulCronTrigger(CronTrigger):
    def __init__(self, *args, **kw):
        self._zuul_jitter = kw.pop('jitter')
        super().__init__(*args, **kw)

    def get_next_fire_time(self, previous_fire_time, now):
        next_time = super().get_next_fire_time(previous_fire_time, now)
        if self._zuul_jitter:
            next_time = next_time + datetime.timedelta(
                seconds=self._zuul_jitter)
            if self.end_date:
                next_time = min(next_time, self.end_date)
        return next_time



# Copyright (C) 2011-2013 OpenStack Foundation
# Copyright 2022, 2024 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import uuid

from zuul.driver.openstack.openstackendpoint import OpenstackProviderEndpoint


class FakeOpenstackObject:
    def __init__(self, **kw):
        self.__dict__.update(kw)
        self.__kw = list(kw.keys())

    def _get_dict(self):
        data = {}
        for k in self.__kw:
            data[k] = getattr(self, k)
        return data

    def __contains__(self, key):
        return key in self.__dict__

    def __getitem__(self, key, default=None):
        return getattr(self, key, default)

    def __setitem__(self, key, value):
        setattr(self, key, value)

    def get(self, key, default=None):
        return getattr(self, key, default)

    def set(self, key, value):
        setattr(self, key, value)


class FakeOpenstackFlavor(FakeOpenstackObject):
    pass


class FakeOpenstackServer(FakeOpenstackObject):
    pass


class FakeOpenstackLocation(FakeOpenstackObject):
    pass


class FakeOpenstackImage(FakeOpenstackObject):
    pass


class FakeOpenstackNetwork(FakeOpenstackObject):
    pass


class FakeOpenstackFloatingIp(FakeOpenstackObject):
    def _fake_toDict(self):
        return {
            'version': 4,
            'addr': self.floating_ip_address,
            'OS-EXT-IPS:type': 'floating',
        }


class FakeOpenstackPort(FakeOpenstackObject):
    pass


class FakeOpenstackCloud:
    log = logging.getLogger("zuul.FakeOpenstackCloud")

    def __init__(self,
                 region,
                 needs_floating_ip=False,
                 auto_attach_floating_ip=True,
                 ):
        self._fake_region = region
        self._fake_needs_floating_ip = needs_floating_ip
        self._fake_auto_attach_floating_ip = auto_attach_floating_ip
        self.servers = []
        self.volumes = []
        self.images = []
        self.floating_ips = []
        self.flavors = [
            FakeOpenstackFlavor(
                id='425e3203150e43d6b22792f86752533d',
                name='Fake Flavor',
                ram=8192,
                vcpus=4,
            )
        ]
        self.networks = [
            FakeOpenstackNetwork(
                id=uuid.uuid4().hex,
                name='fake-network',
            )
        ]
        self.ports = []

    def _getConnection(self):
        return FakeOpenstackConnection(self)


class FakeOpenstackResponse:
    def __init__(self, data):
        self._data = data
        self.links = []

    def json(self):
        return self._data


class FakeOpenstackSession:
    def __init__(self, cloud):
        self.cloud = cloud

    def get(self, uri, headers, params):
        if uri == '/servers/detail':
            server_list = []
            for server in self.cloud.servers:
                data = server._get_dict()
                data['status'] = 'ACTIVE'
                data['hostId'] = data.pop('host_id')
                data['OS-EXT-AZ:availability_zone'] = data.pop('location').zone
                data['os-extended-volumes:volumes_attached'] =\
                    data.pop('volumes')
                server_list.append(data)
            return FakeOpenstackResponse({'servers': server_list})


class FakeOpenstackConfig:
    def __init__(self, region):
        self.config = {}
        self.config['image_format'] = 'qcow2'
        self.config['region_name'] = region


class FakeOpenstackConnection:
    log = logging.getLogger("zuul.FakeOpenstackConnection")

    def __init__(self, cloud):
        self.cloud = cloud
        self.compute = FakeOpenstackSession(cloud)
        self.config = FakeOpenstackConfig(cloud._fake_region or 'region1')

    def _needs_floating_ip(self, server, nat_destination):
        return self.cloud._fake_needs_floating_ip

    def _use_neutron_floating(self):
        return self.cloud._fake_needs_floating_ip

    def _has_floating_ips(self):
        return self.cloud._fake_needs_floating_ip

    def delete_unattached_floating_ips(self):
        for x in self.cloud.floating_ips:
            if not getattr(x, 'port', None):
                self.cloud.floating_ips.remove(x)

    def list_flavors(self, get_extra=False):
        return self.cloud.flavors

    def list_volumes(self):
        return self.cloud.volumes

    def list_ports(self, filters=None):
        if filters and filters.get('status'):
            return [p for p in self.cloud.ports
                    if p.status == filters['status']]
        return self.cloud.ports

    def delete_port(self, port_id):
        for x in self.cloud.ports:
            if x.id == port_id:
                self.cloud.ports.remove(x)
                return

    def get_network(self, name_or_id, filters=None):
        for x in self.cloud.networks:
            if x.id == name_or_id or x.name == name_or_id:
                return x

    def list_servers(self):
        return self.cloud.servers

    def create_server(self, wait=None, name=None, image=None,
                      flavor=None, config_drive=None, key_name=None,
                      nics=None, meta=None, availability_zone=None):
        location = FakeOpenstackLocation(zone=availability_zone)
        if self.cloud._fake_needs_floating_ip:
            addresses = dict(
                public=[],
                private=[dict(version=4, addr='198.51.100.1')]
            )
            interface_ip = '198.51.100.1'
        else:
            addresses = dict(
                public=[dict(version=4, addr='198.51.100.1'),
                        dict(version=6, addr='2001:db8::1')],
                private=[dict(version=4, addr='198.51.100.1')]
            )
            interface_ip = '198.51.100.1'

        args = dict(
            id=uuid.uuid4().hex,
            name=name,
            host_id='fake_host_id',
            location=location,
            volumes=[],
            status='BUILDING',
            addresses=addresses,
            interface_ip=interface_ip,
            flavor=flavor,
            metadata=meta,
        )
        server = FakeOpenstackServer(**args)
        self.cloud.servers.append(server)
        return server

    def delete_server(self, name_or_id):
        for x in self.cloud.servers:
            if x.id == name_or_id:
                self.cloud.servers.remove(x)
                return

    def create_image(self, wait=None, name=None, filename=None,
                     is_public=None, md5=None, sha256=None,
                     timeout=None, **meta):
        image = FakeOpenstackImage(
            id=uuid.uuid4().hex,
            name=name,
            filename=filename,
            is_public=is_public,
            md5=md5,
            sha256=sha256,
            status='ACTIVE',
        )
        self.cloud.images.append(image)
        return image

    def delete_image(self, name_or_id):
        for x in self.cloud.servers:
            if x.id == name_or_id:
                self.cloud.servers.remove(x)
                return

    def create_floating_ip(self, server, wait=None):
        args = dict(
            id=uuid.uuid4().hex,
            floating_ip_address='fake',
            status='ACTIVE',
        )
        if self.cloud._fake_auto_attach_floating_ip:
            args['port_id'] = 'fake'
        fip = FakeOpenstackFloatingIp(**args)
        self.cloud.floating_ips.append(fip)
        if self.cloud._fake_auto_attach_floating_ip:
            server['addresses']['public'].append(fip._fake_toDict())
        return fip

    def list_floating_ips(self):
        return self.cloud.floating_ips

    def get_floating_ip(self, id, filters=None):
        for x in self.cloud.floating_ips:
            if x.id == id:
                return x
            for k, v in filters.items():
                if x[k] == v:
                    return x

    def delete_floating_ip(self, floating_ip_id, retry=1):
        for x in self.cloud.floating_ips:
            if x.id == floating_ip_id:
                self.cloud.floating_ips.remove(x)
                return

    def _attach_ip_to_server(self, server, floating_ip, skip_attach=False):
        if floating_ip.get('port_id'):
            raise Exception("Attaching already attached fip")
        floating_ip['port_id'] = 'fake'
        server['addresses']['public'].append(floating_ip._fake_toDict())

    def get_compute_limits(self):
        return FakeOpenstackObject(
            max_total_cores=self.cloud.max_cores,
            max_total_instances=self.cloud.max_instances,
            max_total_ram_size=self.cloud.max_ram,
            total_cores_used=4 * len(self.cloud.servers),
            total_instances_used=len(self.cloud.servers),
            total_ram_used=8192 * len(self.cloud.servers),
        )

    def get_volume_limits(self):
        return FakeOpenstackObject(
            absolute=dict(
                maxTotalVolumes=self.cloud.max_volumes,
                maxTotalVolumeGigabytes=self.cloud.max_volume_gb,
            ))


class FakeOpenstackProviderEndpoint(OpenstackProviderEndpoint):
    def _getRegionConfig(self):
        return FakeOpenstackConfig('region1')

    def _getClient(self):
        return self._fake_cloud[self.region]._getConnection()

    def _expandServer(self, server):
        return server



"""Change patchset to string

Revision ID: 19d3a3ebfe1d
Revises: cfc0dc45f341
Create Date: 2018-01-10 07:42:16.546751

"""

# revision identifiers, used by Alembic.
revision = '19d3a3ebfe1d'
down_revision = 'cfc0dc45f341'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa

BUILDSET_TABLE = 'zuul_buildset'


def upgrade(table_prefix=''):
    op.alter_column(table_prefix + BUILDSET_TABLE,
                    'patchset',
                    type_=sa.String(255),
                    existing_nullable=True)


def downgrade():
    raise Exception("Downgrades not supported")



# Copyright 2019 BMW Group
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging
import textwrap
import zuul.cmd
from zuul.lib.ansible import AnsibleManager


class ManageAnsible(zuul.cmd.ZuulApp):

    app_name = 'manage-ansible'
    app_description = textwrap.dedent('''
        Zuul ansible manager.

        This command installs or upgrades all supported Ansible installations
        so zuul can use them.

        You can set the following environnment variables
        to install additional packages you might need along with ansible.
        These variables must contain a space separated list of dependencies
        that can be parsed by pip.

        ANSIBLE_EXTRA_PACKAGES
            Packages to add to every ansible installation.

        ANSIBLE_<VERSION>_EXTRA_PACKAGES
            Packages to add to a specific version of Ansible. The version must
            be the same as listed in 'zuul-manage-ansible -l' but without
            special characters. e.g. ANSIBLE_27_EXTRA_PACKAGES=myextradep
        ''')

    log = logging.getLogger('zuul.ManageAnsible')

    def createParser(self):
        parser = super().createParser()
        parser.add_argument('-v', dest='verbose', action='store_true',
                            help='verbose output')
        parser.add_argument('-u', dest='upgrade', action='store_true',
                            help='upgrade ansible versions')
        parser.add_argument('-l', dest='list_supported', action='store_true',
                            help='list supported versions')
        parser.add_argument('--validate', dest='validate',
                            action='store_true',
                            help='validate installed versions')
        parser.add_argument('-r', dest='install_root', default=None,
                            help='root path for ansible venv installations')
        return parser

    def _setup_logging(self):
        """Manage ansible logging does not rely on conf file"""
        if self.args.verbose:
            logging.basicConfig(level=logging.DEBUG)
        else:
            logging.basicConfig(level=logging.INFO)

    def main(self):
        self.parseArguments()
        try:
            self.readConfig()
        except Exception:
            # This script must be able to run without config so this can be
            # safely ignored here.
            pass
        self._setup_logging()

        manager = AnsibleManager(runtime_install_root=self.args.install_root)

        if self.args.list_supported:
            versions = []
            for version, default in manager.getSupportedVersions():
                if default:
                    version = version + ' (default)'
                versions.append(version)
            print('\n'.join(versions))
            return

        if self.args.validate:
            manager.validate()
            return

        manager.install(upgrade=self.args.upgrade)


def main():
    ManageAnsible().main()



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""add_artifact_table

Revision ID: ea2bae776723
Revises: f181b33958c6
Create Date: 2018-11-26 14:48:54.463512

"""

# revision identifiers, used by Alembic.
revision = 'ea2bae776723'
down_revision = 'f181b33958c6'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa


ARTIFACT_TABLE = 'zuul_artifact'
BUILD_TABLE = 'zuul_build'


def upgrade(table_prefix=''):
    op.create_table(
        table_prefix + ARTIFACT_TABLE,
        sa.Column('id', sa.Integer, primary_key=True),
        sa.Column('build_id', sa.Integer,
                  sa.ForeignKey(table_prefix + BUILD_TABLE + ".id")),
        sa.Column('name', sa.String(255)),
        sa.Column('url', sa.TEXT()),
    )


def downgrade():
    raise Exception("Downgrades not supported")



# Copyright 2024 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import configparser
import logging
import os

from zuul.connection import BaseConnection


class AwsConnection(BaseConnection):
    driver_name = 'aws'
    log = logging.getLogger("zuul.AwsConnection")

    def __init__(self, driver, connection_name, connection_config):
        super().__init__(driver, connection_name, connection_config)

        # Users can provide credentials directly in zuul.conf, or in the
        # AWS shared credentials file.  We will check those places in
        # that order.
        self.shared_credentials_file = self.connection_config.get(
            'shared_credentials_file')

        self.access_key_id = self.connection_config.get(
            'access_key_id')
        self.secret_access_key = self.connection_config.get(
            'secret_access_key')
        self.profile = self.connection_config.get(
            'profile')
        # Rate limit: requests/second
        self.rate = self.connection_config.get('rate', 2)

        if (not self.access_key_id) and self.shared_credentials_file:
            path = os.path.expanduser(self.shared_credentials_file)
            config = configparser.ConfigParser()
            config.read(path)

            section = config[self.profile]
            self.access_key_id = section['access_key_id']
            self.secret_access_key = section['secret_access_key']



# Copyright 2015 Hewlett-Packard Development Company, L.P.
# Copyright 2017 IBM Corp.
# Copyright 2017 Red Hat, Inc.
# Copyright 2023 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import copy
import re
import re2
import time

from zuul.model import Change, TriggerEvent, EventFilter, RefFilter
from zuul.model import FalseWithReason
from zuul.driver.util import time_to_seconds, to_list

EMPTY_GIT_REF = '0' * 40  # git sha of all zeros, used during creates/deletes


class PullRequest(Change):
    def __init__(self, project):
        super(PullRequest, self).__init__(project)
        self.pr = None
        self.updated_at = None
        self.title = None
        self.body_text = None
        self.reviews = []
        self.files = []
        self.labels = []
        self.draft = None
        self.mergeable = True
        self.review_decision = None
        self.unresolved_conversations = None
        self.required_contexts = set()
        self.contexts = set()
        self.branch_protected = False
        self.merge_commit_sha = None

    @property
    def status(self):
        return ["{}:{}:{}".format(*c) for c in self.contexts]

    @property
    def successful_contexts(self) -> set:
        if not self.contexts:
            return set()
        return set(
            s[1] for s in self.contexts if s[2] == 'success'
        )

    def isUpdateOf(self, other):
        if (self.project == other.project and
            hasattr(other, 'number') and self.number == other.number and
            hasattr(other, 'patchset') and self.patchset != other.patchset and
            hasattr(other, 'updated_at') and
            self.updated_at > other.updated_at):
            return True
        return False

    def serialize(self):
        d = super().serialize()
        d.update({
            "pr": self.pr,
            "updated_at": self.updated_at,
            "title": self.title,
            "body_text": self.body_text,
            "reviews": list(self.reviews),
            "labels": self.labels,
            "draft": self.draft,
            "mergeable": self.mergeable,
            "review_decision": self.review_decision,
            "unresolved_conversations": self.unresolved_conversations,
            "required_contexts": list(self.required_contexts),
            "contexts": list(self.contexts),
            "branch_protected": self.branch_protected,
        })
        return d

    def deserialize(self, data):
        super().deserialize(data)
        self.pr = data.get("pr")
        self.updated_at = data.get("updated_at")
        self.title = data.get("title")
        self.body_text = data.get("body_text")
        self.reviews = data.get("reviews", [])
        self.labels = data.get("labels", [])
        self.draft = data.get("draft")
        self.mergeable = data.get("mergeable", True)
        self.review_decision = data.get("review_decision")
        self.unresolved_conversations = data.get("unresolved_conversations")
        self.required_contexts = set(data.get("required_contexts", []))
        self.contexts = set(tuple(c) for c in data.get("contexts", []))
        self.branch_protected = data.get("branch_protected", False)


class GithubTriggerEvent(TriggerEvent):
    def __init__(self):
        super(GithubTriggerEvent, self).__init__()
        self.title = None
        self.label = None
        self.unlabel = None
        self.action = None
        self.delivery = None
        self.check_run = None
        self.status = None
        self.commits = []
        self.body_edited = None
        self.branch_protection_changed = None
        self.default_branch_changed = None

    def toDict(self):
        d = super().toDict()
        d["title"] = self.title
        d["label"] = self.label
        d["unlabel"] = self.unlabel
        d["action"] = self.action
        d["delivery"] = self.delivery
        d["check_run"] = self.check_run
        d["status"] = self.status
        d["commits"] = self.commits
        d["body_edited"] = self.body_edited
        d["branch_protection_changed"] = self.branch_protection_changed
        d["default_branch_changed"] = self.default_branch_changed
        return d

    def updateFromDict(self, d):
        super().updateFromDict(d)
        self.title = d["title"]
        self.label = d["label"]
        self.unlabel = d["unlabel"]
        self.action = d["action"]
        self.delivery = d["delivery"]
        self.check_run = d["check_run"]
        self.status = d["status"]
        self.commits = d["commits"]
        self.body_edited = d["body_edited"]
        self.branch_protection_changed = d.get("branch_protection_changed")
        self.default_branch_changed = d.get("default_branch_changed")

    def isBranchProtectionChanged(self):
        return bool(self.branch_protection_changed)

    def isDefaultBranchChanged(self):
        return bool(self.default_branch_changed)

    def isPatchsetCreated(self):
        if self.type == 'pull_request':
            return self.action in ['opened', 'changed']
        return False

    def isMessageChanged(self):
        return bool(self.body_edited)

    def isChangeAbandoned(self):
        if self.type == 'pull_request':
            return 'closed' == self.action
        return False

    def _repr(self):
        r = [super(GithubTriggerEvent, self)._repr()]
        if self.action:
            r.append(self.action)
        r.append(self.canonical_project_name)
        if self.change_number:
            r.append('%s,%s' % (self.change_number, self.patch_number))
        if self.delivery:
            r.append('delivery: %s' % self.delivery)
        if self.check_run:
            r.append('check_run: %s' % self.check_run)
        return ' '.join(r)


class GithubEventFilter(EventFilter):
    def __init__(self, connection_name, trigger, types=[],
                 branches=[], refs=[], comments=[], actions=[],
                 labels=[], unlabels=[], states=[], statuses=[],
                 required_statuses=[], check_runs=[],
                 ignore_deletes=True,
                 require=None, reject=None, debug=None):

        EventFilter.__init__(self, connection_name, trigger, debug)

        # TODO: Backwards compat, remove after 9.x:
        if required_statuses and require is None:
            require = {'status': required_statuses}

        if require:
            self.require_filter = GithubRefFilter.requiresFromConfig(
                connection_name, require)
        else:
            self.require_filter = None

        if reject:
            self.reject_filter = GithubRefFilter.rejectFromConfig(
                connection_name, reject)
        else:
            self.reject_filter = None

        self._types = [x.pattern for x in types]
        self._branches = [x.pattern for x in branches]
        self._refs = [x.pattern for x in refs]
        self._comments = [x.pattern for x in comments]
        self.types = types
        self.branches = branches
        self.refs = refs
        self.comments = comments
        self.actions = actions
        self.labels = labels
        self.unlabels = unlabels
        self.states = states
        self.statuses = statuses
        self.check_runs = check_runs
        self.ignore_deletes = ignore_deletes

    def __repr__(self):
        ret = '<GithubEventFilter'
        ret += ' connection: %s' % self.connection_name
        if self._types:
            ret += ' types: %s' % ', '.join(self._types)
        if self._branches:
            ret += ' branches: %s' % ', '.join(self._branches)
        if self._refs:
            ret += ' refs: %s' % ', '.join(self._refs)
        if self.ignore_deletes:
            ret += ' ignore_deletes: %s' % self.ignore_deletes
        if self._comments:
            ret += ' comments: %s' % ', '.join(self._comments)
        if self.actions:
            ret += ' actions: %s' % ', '.join(self.actions)
        if self.check_runs:
            ret += ' check_runs: %s' % ','.join(self.check_runs)
        if self.labels:
            ret += ' labels: %s' % ', '.join(self.labels)
        if self.unlabels:
            ret += ' unlabels: %s' % ', '.join(self.unlabels)
        if self.states:
            ret += ' states: %s' % ', '.join(self.states)
        if self.statuses:
            ret += ' statuses: %s' % ', '.join(self.statuses)
        if self.require_filter:
            ret += ' require: %s' % repr(self.require_filter)
        if self.reject_filter:
            ret += ' reject: %s' % repr(self.reject_filter)
        ret += '>'

        return ret

    def matches(self, event, change):
        if not super().matches(event, change):
            return False

        # event types are ORed
        matches_type = False
        for etype in self.types:
            if etype.match(event.type):
                matches_type = True
        if self.types and not matches_type:
            return FalseWithReason("Types %s do not match %s" % (
                self.types, event.type))

        # branches are ORed
        matches_branch = False
        for branch in self.branches:
            if branch.match(event.branch):
                matches_branch = True
        if self.branches and not matches_branch:
            return FalseWithReason("Branches %s do not match %s" % (
                self.branches, event.branch))

        # refs are ORed
        matches_ref = False
        if event.ref is not None:
            for ref in self.refs:
                if ref.match(event.ref):
                    matches_ref = True
        if self.refs and not matches_ref:
            return FalseWithReason(
                "Refs %s do not match %s" % (self.refs, event.ref))
        if self.ignore_deletes and event.newrev == EMPTY_GIT_REF:
            # If the updated ref has an empty git sha (all 0s),
            # then the ref is being deleted
            return FalseWithReason("Ref deletion events are ignored")

        # comments are ORed
        matches_comment_re = False
        for comment_re in self.comments:
            if (event.comment is not None and
                comment_re.search(event.comment)):
                matches_comment_re = True
        if self.comments and not matches_comment_re:
            return FalseWithReason("Comments %s do not match %s" % (
                self.comments, event.comment))

        # actions are ORed
        matches_action = False
        for action in self.actions:
            if (event.action == action):
                matches_action = True
        if self.actions and not matches_action:
            return FalseWithReason("Actions %s do not match %s" % (
                self.actions, event.action))

        # check_runs are ORed
        if self.check_runs:
            check_run_found = False
            for check_run in self.check_runs:
                # TODO: construct as ZuulRegex in initializer when re2
                # migration is complete.
                if re2.fullmatch(check_run, event.check_run):
                    check_run_found = True
                    break
            if not check_run_found:
                return FalseWithReason("Check runs %s do not match %s" % (
                    self.check_runs, event.check_run))

        # labels are ORed
        if self.labels and event.label not in self.labels:
            return FalseWithReason("Labels %s do not match %s" % (
                self.labels, event.label))

        # unlabels are ORed
        if self.unlabels and event.unlabel not in self.unlabels:
            return FalseWithReason("Unlabels %s do not match %s" % (
                self.unlabels, event.unlabel))

        # states are ORed
        if self.states and event.state not in self.states:
            return FalseWithReason("States %s do not match %s" % (
                self.states, event.state))

        # statuses are ORed
        if self.statuses:
            status_found = False
            for status in self.statuses:
                # TODO: construct as ZuulRegex in initializer when re2
                # migration is complete.
                if re2.fullmatch(status, event.status):
                    status_found = True
                    break
            if not status_found:
                return FalseWithReason("Statuses %s do not match %s" % (
                    self.statuses, event.status))

        if self.require_filter:
            require_filter_result = self.require_filter.matches(change)
            if not require_filter_result:
                return require_filter_result

        if self.reject_filter:
            reject_filter_result = self.reject_filter.matches(change)
            if not reject_filter_result:
                return reject_filter_result

        return True


class GithubRefFilter(RefFilter):
    def __init__(self, connection_name, statuses=[],
                 reviews=[], reject_reviews=[], open=None,
                 merged=None, current_patchset=None, draft=None,
                 reject_open=None, reject_merged=None,
                 reject_current_patchset=None, reject_draft=None,
                 labels=[], reject_labels=[], reject_statuses=[]):
        RefFilter.__init__(self, connection_name)

        self._required_reviews = copy.deepcopy(reviews)
        self._reject_reviews = copy.deepcopy(reject_reviews)
        self.required_reviews = self._tidy_reviews(self._required_reviews)
        self.reject_reviews = self._tidy_reviews(self._reject_reviews)
        self.required_statuses = statuses
        self.reject_statuses = reject_statuses
        self.required_labels = labels
        self.reject_labels = reject_labels

        if reject_open is not None:
            self.open = not reject_open
        else:
            self.open = open
        if reject_merged is not None:
            self.merged = not reject_merged
        else:
            self.merged = merged
        if reject_current_patchset is not None:
            self.current_patchset = not reject_current_patchset
        else:
            self.current_patchset = current_patchset
        if reject_draft is not None:
            self.draft = not reject_draft
        else:
            self.draft = draft

    @classmethod
    def requiresFromConfig(cls, connection_name, config):
        return cls(
            connection_name=connection_name,
            statuses=to_list(config.get('status')),
            reviews=to_list(config.get('review')),
            labels=to_list(config.get('label')),
            open=config.get('open'),
            merged=config.get('merged'),
            current_patchset=config.get('current-patchset'),
            draft=config.get('draft'),
        )

    @classmethod
    def rejectFromConfig(cls, connection_name, config):
        return cls(
            connection_name=connection_name,
            reject_statuses=to_list(config.get('status')),
            reject_reviews=to_list(config.get('review')),
            reject_labels=to_list(config.get('label')),
            reject_open=config.get('open'),
            reject_merged=config.get('merged'),
            reject_current_patchset=config.get('current-patchset'),
            reject_draft=config.get('draft'),
        )

    def __repr__(self):
        ret = '<GithubRefFilter'

        ret += ' connection_name: %s' % self.connection_name
        if self.required_statuses:
            ret += ' status: %s' % str(self.required_statuses)
        if self.reject_statuses:
            ret += ' reject-status: %s' % str(self.reject_statuses)
        if self.required_reviews:
            ret += (' reviews: %s' %
                    str(self.required_reviews))
        if self.reject_reviews:
            ret += (' reject-reviews: %s' %
                    str(self.reject_reviews))
        if self.required_labels:
            ret += ' labels: %s' % str(self.required_labels)
        if self.reject_labels:
            ret += ' reject-labels: %s' % str(self.reject_labels)
        if self.open is not None:
            ret += ' open: %s' % self.open
        if self.merged is not None:
            ret += ' merged: %s' % self.merged
        if self.current_patchset is not None:
            ret += ' current-patchset: %s' % self.current_patchset
        if self.draft is not None:
            ret += ' draft: %s' % self.draft

        ret += '>'

        return ret

    def _tidy_reviews(self, reviews):
        for r in reviews:
            for k, v in r.items():
                if k == 'username':
                    r['username'] = re.compile(v)
                elif k == 'email':
                    r['email'] = re.compile(v)
                elif k == 'newer-than':
                    r[k] = time_to_seconds(v)
                elif k == 'older-than':
                    r[k] = time_to_seconds(v)
        return reviews

    def _match_review_required_review(self, rreview, review):
        # Check if the required review and review match
        now = time.time()
        by = review.get('by', {})
        for k, v in rreview.items():
            if k == 'username':
                if (not v.search(by.get('username', ''))):
                    return False
            elif k == 'email':
                if (not v.search(by.get('email', ''))):
                    return False
            elif k == 'newer-than':
                t = now - v
                if (review['grantedOn'] < t):
                    return False
            elif k == 'older-than':
                t = now - v
                if (review['grantedOn'] >= t):
                    return False
            elif k == 'type':
                if review['type'] != v:
                    return False
            elif k == 'permission':
                # If permission is read, we've matched. You must have read
                # to provide a review.
                if v != 'read':
                    # Admins have implicit write.
                    if v == 'write':
                        if review['permission'] not in ('write', 'admin'):
                            return False
                    elif v == 'admin':
                        if review['permission'] != 'admin':
                            return False
        return True

    def matchesReviews(self, change):
        if self.required_reviews or self.reject_reviews:
            if not hasattr(change, 'number'):
                # not a PR, no reviews
                return FalseWithReason("Change is not a PR")
            if self.required_reviews and not change.reviews:
                # No reviews means no matching of required bits
                # having reject reviews but no reviews on the change is okay
                return FalseWithReason("Reviews %s do not match %s" % (
                    self.required_reviews, change.reviews))

        return (self.matchesRequiredReviews(change) and
                self.matchesNoRejectReviews(change))

    def matchesRequiredReviews(self, change):
        for rreview in self.required_reviews:
            matches_review = False
            for review in change.reviews:
                if self._match_review_required_review(rreview, review):
                    # Consider matched if any review matches
                    matches_review = True
                    break
            if not matches_review:
                return FalseWithReason(
                    "Required reviews %s do not match %s" % (
                        self.required_reviews, change.reviews))
        return True

    def matchesNoRejectReviews(self, change):
        for rreview in self.reject_reviews:
            for review in change.reviews:
                if self._match_review_required_review(rreview, review):
                    # A review matched, we can reject right away
                    return FalseWithReason("Reject reviews %s match %s" % (
                        self.reject_reviews, change.reviews))
        return True

    def matchesStatuses(self, change):
        if self.required_statuses or self.reject_statuses:
            if not hasattr(change, 'number'):
                # not a PR, no status
                return FalseWithReason("Can not match statuses without PR")
            if self.required_statuses and not change.status:
                return FalseWithReason(
                    "Required statuses %s do not match %s" % (
                        self.required_statuses, change.status))
        required_statuses_results = self.matchesRequiredStatuses(change)
        if not required_statuses_results:
            return required_statuses_results
        return self.matchesNoRejectStatuses(change)

    def matchesRequiredStatuses(self, change):
        # statuses are ORed
        # A PR head can have multiple statuses on it. If the change
        # statuses and the filter statuses are a null intersection, there
        # are no matches and we return false
        if self.required_statuses:
            for required_status in self.required_statuses:
                for status in change.status:
                    # TODO: construct as ZuulRegex in initializer when
                    # re2 migration is complete.
                    if re2.fullmatch(required_status, status):
                        return True
            return FalseWithReason("Required statuses %s do not match %s" % (
                self.required_statuses, change.status))
        return True

    def matchesNoRejectStatuses(self, change):
        # statuses are ANDed
        # If any of the rejected statusses are present, we return false
        for rstatus in self.reject_statuses:
            for status in change.status:
                # TODO: construct as ZuulRegex in initializer when re2
                # migration is complete.
                if re2.fullmatch(rstatus, status):
                    return FalseWithReason("Reject statuses %s match %s" % (
                        self.reject_statuses, change.status))
        return True

    def matchesLabels(self, change):
        if self.required_labels or self.reject_labels:
            if not hasattr(change, 'number'):
                # not a PR, no label
                return FalseWithReason("Can not match labels without PR")
            if self.required_labels and not change.labels:
                # No labels means no matching of required bits
                # having reject labels but no labels on the change is okay
                return FalseWithReason(
                    "Required labels %s does not match %s" % (
                        self.required_labels, change.labels))
        return (self.matchesRequiredLabels(change) and
                self.matchesNoRejectLabels(change))

    def matchesRequiredLabels(self, change):
        for label in self.required_labels:
            if label not in change.labels:
                return FalseWithReason("Labels %s do not match %s" % (
                    self.required_labels, change.labels))
        return True

    def matchesNoRejectLabels(self, change):
        for label in self.reject_labels:
            if label in change.labels:
                return FalseWithReason("Reject labels %s match %s" % (
                    self.reject_labels, change.labels))
        return True

    def matches(self, change):
        statuses_result = self.matchesStatuses(change)
        if not statuses_result:
            return statuses_result

        reviews_result = self.matchesReviews(change)
        if not reviews_result:
            return reviews_result

        labels_result = self.matchesLabels(change)
        if not labels_result:
            return labels_result

        if self.open is not None:
            # if a "change" has no number, it's not a change, but a push
            # and cannot possibly pass this test.
            if hasattr(change, 'number'):
                if self.open != change.open:
                    return FalseWithReason(
                        "Change does not match open requirement")
            else:
                return FalseWithReason("Change is not a PR")

        if self.merged is not None:
            # if a "change" has no number, it's not a change, but a push
            # and cannot possibly pass this test.
            if hasattr(change, 'number'):
                if self.merged != change.is_merged:
                    return FalseWithReason(
                        "Change does not match merged requirement")
            else:
                return FalseWithReason("Change is not a PR")

        if self.current_patchset is not None:
            # if a "change" has no number, it's not a change, but a push
            # and cannot possibly pass this test.
            if hasattr(change, 'number'):
                if self.current_patchset != change.is_current_patchset:
                    return FalseWithReason(
                        "Change does not match current-patchset requirement")
            else:
                return FalseWithReason("Change is not a PR")

        if self.draft is not None:
            # if a "change" has no number, it's not a change, but a push
            # and cannot possibly pass this test.
            if hasattr(change, 'number'):
                if self.draft != change.draft:
                    return FalseWithReason(
                        "Change does not match draft requirement")
            else:
                return FalseWithReason("Change is not a PR")

        return True



# Copyright 2024 BMW Group
# Copyright 2024 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from zuul import model
from zuul.provider import statemachine


class OpenstackProviderNode(model.ProviderNode, subclass_id="openstack"):
    def __init__(self):
        super().__init__()
        self._set(
            openstack_server_id=None,
            openstack_floating_ip_id=None,
        )

    def getDriverData(self):
        return dict(
            openstack_server_id=self.openstack_server_id,
            openstack_floating_ip_id=self.openstack_floating_ip_id,
        )


class OpenstackInstance(statemachine.Instance):
    def __init__(self, cloud, region, server, quota):
        super().__init__()
        self.openstack_server_id = server['id']
        self.metadata = server.get('metadata', {})
        self.private_ipv4 = server.get('private_v4')
        self.private_ipv6 = None
        self.public_ipv4 = server.get('public_v4')
        self.public_ipv6 = server.get('public_v6')
        self.host_id = server['hostId']
        self.cloud = cloud
        self.region = region
        self.az = server.get('OS-EXT-AZ:availability_zone')
        self.interface_ip = server.get('interface_ip')
        self.quota = quota

    def getQuotaInformation(self):
        return self.quota

    @property
    def external_id(self):
        return f'server={self.openstack_server_id}'


class OpenstackResource(statemachine.Resource):
    TYPE_HOST = 'host'
    TYPE_INSTANCE = 'instance'
    TYPE_AMI = 'ami'
    TYPE_SNAPSHOT = 'snapshot'
    TYPE_VOLUME = 'volume'
    TYPE_OBJECT = 'object'

    def __init__(self, metadata, type, id):
        super().__init__(metadata, type)
        self.id = id

    @property
    def unique_id(self):
        return '-'.join([self.type, self.id])



# Copyright 2019 BMW Group
# Copyright 2024 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from graphene import (
    Boolean,
    Field,
    Int,
    Interface,
    List,
    ObjectType,
    Schema,
    String,
)


class ID(String):
    """Github global object ids are strings"""
    pass


class Node(Interface):
    id = ID(required=True)

    @classmethod
    def resolve_type(cls, instance, info):
        kind = getattr(instance, '_graphene_type', None)
        if kind == 'BranchProtectionRule':
            return BranchProtectionRule
        elif kind == 'Commit':
            return Commit
        elif kind == 'CheckSuite':
            return CheckSuite
        elif kind == 'PullRequest':
            return PullRequest


class PageInfo(ObjectType):
    endCursor = String()
    hasPreviousPage = Boolean()
    hasNextPage = Boolean()

    def resolve_endCursor(parent, info):
        return str(parent['after'] + parent['first'])

    def resolve_hasPreviousPage(parent, info):
        return parent['after'] > 0

    def resolve_hasNextPage(parent, info):
        return parent['after'] + parent['first'] < parent['length']


class MatchingRef(ObjectType):
    name = String()

    def resolve_name(parent, info):
        return parent


class MatchingRefs(ObjectType):
    pageInfo = Field(PageInfo)
    nodes = List(MatchingRef)

    def resolve_nodes(parent, info):
        return parent['nodes']

    def resolve_pageInfo(parent, info):
        return parent


class BranchProtectionRule(ObjectType):

    class Meta:
        interfaces = (Node, )

    id = ID(required=True)
    pattern = String()
    requiredStatusCheckContexts = List(String)
    requiresApprovingReviews = Boolean()
    requiresConversationResolution = Boolean()
    requiresCodeOwnerReviews = Boolean()
    matchingRefs = Field(MatchingRefs, first=Int(), after=String(),
                         query=String())
    lockBranch = Boolean()

    def resolve_id(parent, info):
        return parent.id

    def resolve_pattern(parent, info):
        return parent.pattern

    def resolve_requiredStatusCheckContexts(parent, info):
        return parent.required_contexts

    def resolve_requiresApprovingReviews(parent, info):
        return parent.require_reviews

    def resolve_requiresConversationResolution(parent, info):
        return parent.require_conversation_resolution

    def resolve_requiresCodeOwnerReviews(parent, info):
        return parent.require_codeowners_review

    def resolve_lockBranch(parent, info):
        return parent.lock_branch

    def resolve_matchingRefs(parent, info, first, after=None, query=None):
        if after is None:
            after = '0'
        after = int(after)
        values = parent.matching_refs
        if query:
            values = [v for v in values if v == query]
        return dict(
            length=len(values),
            nodes=values[after:after + first],
            first=first,
            after=after,
        )


class BranchProtectionRules(ObjectType):
    pageInfo = Field(PageInfo)
    nodes = List(BranchProtectionRule)

    def resolve_nodes(parent, info):
        return parent['nodes']

    def resolve_pageInfo(parent, info):
        return parent


class Actor(ObjectType):
    login = String()


class StatusContext(ObjectType):
    state = String()
    context = String()
    creator = Field(Actor)

    def resolve_state(parent, info):
        state = parent.state.upper()
        return state

    def resolve_context(parent, info):
        return parent.context

    def resolve_creator(parent, info):
        return parent.creator


class Status(ObjectType):
    contexts = List(StatusContext)

    def resolve_contexts(parent, info):
        return parent


class CheckRun(ObjectType):

    class Meta:
        interfaces = (Node, )

    id = ID(required=True)
    name = String()
    conclusion = String()

    def resolve_name(parent, info):
        return parent.name

    def resolve_conclusion(parent, info):
        if parent.conclusion:
            return parent.conclusion.upper()
        return None


class CheckRuns(ObjectType):
    nodes = List(CheckRun)
    pageInfo = Field(PageInfo)

    def resolve_nodes(parent, info):
        return parent['nodes']

    def resolve_pageInfo(parent, info):
        return parent


class App(ObjectType):
    slug = String()
    name = String()


class CheckSuite(ObjectType):

    class Meta:
        interfaces = (Node, )

    id = ID(required=True)
    app = Field(App)
    checkRuns = Field(CheckRuns, first=Int(), after=String())

    def resolve_app(parent, info):
        if not parent.runs:
            return None
        return parent.runs[0].app

    def resolve_checkRuns(parent, info, first, after=None):
        # Github will only return the single most recent check run for
        # a given name (this is true for REST or graphql), despite the
        # format of the result being a list.
        # Since the check runs are ordered from latest to oldest result we
        # need to traverse the list in reverse order.
        check_runs_by_name = {
            "{}:{}".format(cr.app.name, cr.name):
            cr for cr in reversed(parent.runs)
        }
        if after is None:
            after = '0'
        after = int(after)
        values = list(check_runs_by_name.values())
        return dict(
            length=len(values),
            nodes=values[after:after + first],
            first=first,
            after=after,
        )


class CheckSuites(ObjectType):

    nodes = List(CheckSuite)
    pageInfo = Field(PageInfo)

    def resolve_nodes(parent, info):
        return parent['nodes']

    def resolve_pageInfo(parent, info):
        return parent


class Commit(ObjectType):

    class Meta:
        interfaces = (Node, )

    id = ID(required=True)
    status = Field(Status)
    checkSuites = Field(CheckSuites, first=Int(), after=String())

    def resolve_status(parent, info):
        seen = set()
        result = []
        for status in parent._statuses:
            if status.context not in seen:
                seen.add(status.context)
                result.append(status)
        # Github returns None if there are no results
        return result or None

    def resolve_checkSuites(parent, info, first, after=None):
        if after is None:
            after = '0'
        after = int(after)
        # Each value is a list of check runs for that suite
        values = list(parent._check_suites.values())
        return dict(
            length=len(values),
            nodes=values[after:after + first],
            first=first,
            after=after,
        )


class PullRequestReviewThread(ObjectType):
    isResolved = Boolean()

    def resolve_isResolved(parent, info):
        return parent.resolved


class PullRequestReviewThreads(ObjectType):
    nodes = List(PullRequestReviewThread)
    pageInfo = Field(PageInfo)

    def resolve_nodes(parent, info):
        return parent['nodes']

    def resolve_pageInfo(parent, info):
        return parent


class PullRequest(ObjectType):
    class Meta:
        interfaces = (Node, )

    id = ID(required=True)
    isDraft = Boolean()
    reviewDecision = String()
    mergeable = String()
    merged = Boolean()
    reviewThreads = Field(PullRequestReviewThreads,
                          first=Int(), after=String())

    def resolve_id(parent, info):
        return parent.id

    def resolve_isDraft(parent, info):
        return parent.draft

    def resolve_mergeable(parent, info):
        return "MERGEABLE" if parent.mergeable else "CONFLICTING"

    def resolve_merged(parent, info):
        return parent.is_merged

    def resolve_reviewThreads(parent, info, first, after=None):
        if after is None:
            after = '0'
        after = int(after)
        values = parent.review_threads
        return dict(
            length=len(values),
            nodes=values[after:after + first],
            first=first,
            after=after,
        )

    def resolve_reviewDecision(parent, info):
        if hasattr(info.context, 'version') and info.context.version:
            if info.context.version < (2, 21, 0):
                raise Exception('Field unsupported')

        # Check branch protection rules if reviews are required
        org, project = parent.project.split('/')
        repo = info.context._data.repos[(org, project)]
        rule = repo._branch_protection_rules.get(parent.branch)
        if not rule or not rule.require_reviews:
            # Github returns None if there is no review required
            return None

        approvals = [r for r in parent.reviews
                     if r.data['state'] == 'APPROVED']
        if approvals:
            return 'APPROVED'

        return 'REVIEW_REQUIRED'


class Repository(ObjectType):
    name = String()
    branchProtectionRules = Field(BranchProtectionRules,
                                  first=Int(), after=String())
    pullRequest = Field(PullRequest, number=Int(required=True))
    object = Field(Commit, expression=String(required=True))

    def resolve_name(parent, info):
        org, name = parent.name.split('/')
        return name

    def resolve_branchProtectionRules(parent, info, first, after=None):
        if after is None:
            after = '0'
        after = int(after)
        values = list(parent._branch_protection_rules.values())
        return dict(
            length=len(values),
            nodes=values[after:after + first],
            first=first,
            after=after,
        )

    def resolve_pullRequest(parent, info, number):
        return parent.data.pull_requests.get(number)

    def resolve_object(parent, info, expression):
        return parent._commits.get(expression)


class FakeGithubQuery(ObjectType):
    repository = Field(Repository, owner=String(required=True),
                       name=String(required=True))
    node = Field(Node, id=ID(required=True))

    def resolve_repository(root, info, owner, name):
        return info.context._data.repos.get((owner, name))

    def resolve_node(root, info, id):
        for repo in info.context._data.repos.values():
            for rule in repo._branch_protection_rules.values():
                if rule.id == id:
                    return rule
            for commit in repo._commits.values():
                if commit.id == id:
                    return commit
                for suite in commit._check_suites.values():
                    if suite.id == id:
                        return suite
        for pr in info.context._data.pull_requests.values():
            if pr.id == id:
                return pr


def getGrapheneSchema():
    return Schema(query=FakeGithubQuery, types=[ID])



# Copyright 2017 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


def zuul_legacy_vars(zuul):
    # intentionally omitted:
    # BASE_LOG_PATH
    # JOB_TAGS
    # LOG_PATH
    # ZUUL_COMMIT
    # ZUUL_REF
    # ZUUL_URL
    #
    # newly added to all builds:
    # ZUUL_SHORT_PROJECT_NAME
    #
    # existing in most builds but newly added for periodic:
    # ZUUL_BRANCH

    short_name = zuul['project']['name'].split('/')[-1]
    params = dict(ZUUL_UUID=zuul['build'],
                  ZUUL_PROJECT=zuul['project']['name'],
                  ZUUL_SHORT_PROJECT_NAME=short_name,
                  ZUUL_PIPELINE=zuul['pipeline'],
                  ZUUL_VOTING=zuul['voting'],
                  WORKSPACE='/home/zuul/workspace')
    if 'timeout' in zuul and zuul['timeout'] is not None:
        params['BUILD_TIMEOUT'] = str(int(zuul['timeout']) * 1000)
    if 'branch' in zuul:
        params['ZUUL_BRANCH'] = zuul['branch']

    if 'change' in zuul:
        changes_str = '^'.join(
            ['%s:%s:refs/changes/%s/%s/%s' % (
                i['project']['name'],
                i['branch'],
                str(i['change'])[-2:],
                i['change'],
                i['patchset'])
             for i in zuul['items']])
        params['ZUUL_CHANGES'] = changes_str

        change_ids = ' '.join(['%s,%s' % (i['change'], i['patchset'])
                               for i in zuul['items']])
        params['ZUUL_CHANGE_IDS'] = change_ids
        params['ZUUL_CHANGE'] = str(zuul['change'])
        params['ZUUL_PATCHSET'] = str(zuul['patchset'])

    if 'newrev' in zuul or 'oldrev' in zuul:
        params['ZUUL_REFNAME'] = zuul['ref']
        params['ZUUL_OLDREV'] = zuul.get('oldrev', '0' * 40)
        params['ZUUL_NEWREV'] = zuul.get('newrev', '0' * 40)

    params['TOX_TESTENV_PASSENV'] = ' '.join(params.keys())
    return params


class FilterModule(object):

    def filters(self):
        return {
            'zuul_legacy_vars': zuul_legacy_vars,
        }



# Copyright 2013 Rackspace Australia
# Copyright 2024 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging
import voluptuous as v

from zuul.driver.gerrit.gerritsource import GerritSource
from zuul.driver.gerrit.gerritmodel import GerritChange
from zuul.lib.logutil import get_annotated_logger
from zuul.model import Change
from zuul.reporter import BaseReporter

# commentSizeLimit default set by Gerrit.  Gerrit is a bit
# vague about what this means, it says
#
#  Comments which exceed this size will be rejected ... Size
#  computation is approximate and may be off by roughly 1% ...
#  Default is 16k
#
# This magic number is int((16 << 10) * 0.98).  Robot comments
# are accounted for separately.
GERRIT_HUMAN_MESSAGE_LIMIT = 16056


class GerritReporter(BaseReporter):
    """Sends off reports to Gerrit."""

    name = 'gerrit'
    log = logging.getLogger("zuul.GerritReporter")

    def __init__(self, driver, connection, config=None):
        super(GerritReporter, self).__init__(driver, connection, config)
        action = self.config.copy()
        self._create_comment = action.pop('comment', True)
        self._submit = action.pop('submit', False)
        self._checks_api = action.pop('checks-api', None)
        self._notify = action.pop('notify', None)
        self._labels = {str(k): v for k, v in action.items()}

    def __repr__(self):
        return f"<GerritReporter: {self._action}>"

    def report(self, item, phase1=True, phase2=True):
        """Send a message to gerrit."""
        log = get_annotated_logger(self.log, item.event)

        ret = []
        for change in item.changes:
            err = self._reportChange(item, change, log, phase1, phase2)
            if err:
                ret.append(err)
        return ret

    def _reportChange(self, item, change, log, phase1=True, phase2=True):
        """Send a message to gerrit."""
        # If the source is no GerritSource we cannot report anything here.
        if not isinstance(change.project.source, GerritSource):
            return

        # We can only report changes, not plain branches
        if not isinstance(change, Change):
            return

        # For supporting several Gerrit connections we also must filter by
        # the canonical hostname.
        if change.project.source.connection.canonical_hostname != \
                self.connection.canonical_hostname:
            return

        comments = self.getFileComments(item, change)
        if self._create_comment:
            message = self._formatItemReport(item, change=change)

            b_len = len(message.encode('utf-8'))
            if b_len >= GERRIT_HUMAN_MESSAGE_LIMIT:
                log.info("Message too long, using short form")
                message = self._formatItemReport(item, change=change,
                                                 short=True)
            b_len = len(message.encode('utf-8'))
            if b_len >= GERRIT_HUMAN_MESSAGE_LIMIT:
                log.info("Message truncated %d > %d" %
                         (b_len, GERRIT_HUMAN_MESSAGE_LIMIT))
                message = ("%s... (truncated)" %
                           message[:GERRIT_HUMAN_MESSAGE_LIMIT - 20])
        else:
            message = ''

        log.debug("Report change %s, params %s, message: %s, comments: %s",
                  change, self.config, message, comments)
        if phase2 and self._submit and not hasattr(change, '_ref_sha'):
            # If we're starting to submit a bundle, save the current
            # ref sha for every item in the bundle.
            # Store a dict of project,branch -> sha so that if we have
            # duplicate project/branches, we only query once.
            ref_shas = {}
            for other_change in item.changes:
                if not isinstance(other_change, GerritChange):
                    continue
                key = (other_change.project, other_change.branch)
                ref_sha = ref_shas.get(key)
                if not ref_sha:
                    ref_sha = other_change.project.source.getRefSha(
                        other_change.project,
                        'refs/heads/' + other_change.branch)
                    ref_shas[key] = ref_sha
                other_change._ref_sha = ref_sha

        return self.connection.review(item, change, message,
                                      self._submit, self._labels,
                                      self._checks_api, self._notify,
                                      comments,
                                      phase1, phase2,
                                      zuul_event_id=item.event)

    def getSubmitAllowNeeds(self, manager):
        """Get a list of code review labels that are allowed to be
        "needed" in the submit records for a change, with respect
        to this queue.  In other words, the list of review labels
        this reporter itself is likely to set before submitting.
        """
        return self._labels


def getSchema():
    gerrit_reporter = v.Any(str, v.Schema(dict))
    return gerrit_reporter



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from abc import ABCMeta

from zuul import model
from zuul.lib.logutil import get_annotated_logger
from zuul.manager import PipelineManager, StaticChangeQueueContextManager
from zuul.manager import DynamicChangeQueueContextManager


class ChangeQueueManager:

    def __init__(self, pipeline_manager, name=None, per_branch=False):
        self.log = pipeline_manager.log
        self.pipeline_manager = pipeline_manager
        self.name = name
        self.per_branch = per_branch
        self.projects = []
        self.created_for_branches = {}

    def addProject(self, project):
        self.projects.append(project)

    def getOrCreateQueue(self, project, branch):
        change_queue = self.created_for_branches.get(branch)

        if not change_queue:
            name = self.name or project.name
            change_queue = self.pipeline_manager.constructChangeQueue(name)
            self.pipeline_manager.state.addQueue(change_queue)
            self.created_for_branches[branch] = change_queue

        if not change_queue.matches(project.canonical_name, branch):
            change_queue.addProject(project, branch)
            self.log.debug("Added project %s to queue: %s" %
                           (project, change_queue))

        return change_queue


class SharedQueuePipelineManager(PipelineManager, metaclass=ABCMeta):
    """Intermediate class that adds the shared-queue behavior.

    This is not a full pipeline manager; it just adds the shared-queue
    behavior to the base class and is used by the dependent and serial
    managers.
    """

    changes_merge = False

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.change_queue_managers = []

    def buildChangeQueues(self, layout):
        self.log.debug("Building shared change queues")
        change_queues_managers = {}
        tenant = self.tenant
        layout_project_configs = layout.project_configs

        for project_name, project_configs in layout_project_configs.items():
            (trusted, project) = tenant.getProject(project_name)
            queue_name = None
            project_in_pipeline = False
            for project_config in layout.getAllProjectConfigs(project_name):
                project_pipeline_config = project_config.pipelines.get(
                    self.pipeline.name)
                if not queue_name:
                    queue_name = project_config.queue_name
                if project_pipeline_config is None:
                    continue
                project_in_pipeline = True
            if not project_in_pipeline:
                continue

            # Check if the queue is global or per branch
            queue = layout.queues.get(queue_name)
            per_branch = queue and queue.per_branch

            if queue_name and queue_name in change_queues_managers:
                change_queue_manager = change_queues_managers[queue_name]
            else:
                change_queue_manager = ChangeQueueManager(
                    self, name=queue_name, per_branch=per_branch)
                if queue_name:
                    # If this is a named queue, keep track of it in
                    # case it is referenced again.  Otherwise, it will
                    # have a name automatically generated from its
                    # constituent projects.
                    change_queues_managers[queue_name] = change_queue_manager
                self.change_queue_managers.append(change_queue_manager)
                self.log.debug("Created queue: %s" % change_queue_manager)
            change_queue_manager.addProject(project)
            self.log.debug("Added project %s to queue managers: %s" %
                           (project, change_queue_manager))

    def getChangeQueue(self, change, event, existing=None):
        log = get_annotated_logger(self.log, event)

        # Ignore the existing queue, since we can always get the correct queue
        # from the pipeline. This avoids enqueuing changes in a wrong queue
        # e.g. during re-configuration.
        queue = self.state.getQueue(change.project.canonical_name,
                                    change.branch)
        if queue:
            return StaticChangeQueueContextManager(queue)
        else:
            # Change queues in the dependent pipeline manager are created
            # lazy so first check the managers for the project.
            matching_managers = [t for t in self.change_queue_managers
                                 if change.project in t.projects]
            if matching_managers:
                manager = matching_managers[0]
                branch = None
                if manager.per_branch:
                    # The change queue is not existing yet for this branch
                    branch = change.branch

                # We have a queue manager but no queue yet, so create it
                return StaticChangeQueueContextManager(
                    manager.getOrCreateQueue(change.project, branch)
                )

            # No specific per-branch queue matched so look again with no branch
            queue = self.state.getQueue(change.project.canonical_name, None)
            if queue:
                return StaticChangeQueueContextManager(queue)

            # There is no existing queue for this change. Create a
            # dynamic one for this one change's use
            change_queue = model.ChangeQueue.new(
                self.current_context,
                manager=self,
                dynamic=True)
            change_queue.addProject(change.project, None)
            self.state.addQueue(change_queue)
            log.debug("Dynamically created queue %s", change_queue)
            return DynamicChangeQueueContextManager(
                change_queue, allow_delete=True)



# Copyright 2017 Red Hat, Inc.
# Copyright 2023-2024 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

# Utility methods to promote consistent configuration among drivers.

import concurrent
import os
import threading
import time

import voluptuous as vs

from zuul.configloader import ZUUL_REGEX, make_regex  # noqa


def time_to_seconds(s):
    if s.endswith('s'):
        return int(s[:-1])
    if s.endswith('m'):
        return int(s[:-1]) * 60
    if s.endswith('h'):
        return int(s[:-1]) * 60 * 60
    if s.endswith('d'):
        return int(s[:-1]) * 24 * 60 * 60
    if s.endswith('w'):
        return int(s[:-1]) * 7 * 24 * 60 * 60
    raise Exception("Unable to parse time value: %s" % s)


def scalar_or_list(x):
    return vs.Any([x], x)


def to_list(item):
    if not item:
        return []
    if isinstance(item, list):
        return item
    return [item]


class RateLimitInstance:
    def __init__(self, limiter, logger, msg):
        self.limiter = limiter
        self.logger = logger
        self.msg = msg

    def __enter__(self):
        self.delay = self.limiter._enter()
        self.start_time = time.monotonic()

    def __exit__(self, etype, value, tb):
        end_time = time.monotonic()
        self.limiter._exit(etype, value, tb)
        self.logger("%s in %ss after %ss delay",
                    self.msg,
                    end_time - self.start_time,
                    self.delay)


class RateLimiter:
    """A Rate limiter

    :param str name: The provider name; used in logging.
    :param float rate_limit: The rate limit expressed in
        requests per second.

    Example:
    .. code:: python

        rate_limiter = RateLimiter('provider', 1.0)
        with rate_limiter:
            api_call()

    You can optionally use the limiter as a callable in which case it
    will log a supplied message with timing information.

    .. code:: python

        rate_limiter = RateLimiter('provider', 1.0)
        with rate_limiter(log.debug, "an API call"):
            api_call()

    """

    def __init__(self, name, rate_limit):
        self._running = True
        self.name = name
        if not rate_limit:
            self.delta = 0.0
        else:
            self.delta = 1.0 / rate_limit
        self.last_ts = None
        self.lock = threading.Lock()

    def __call__(self, logmethod, msg):
        return RateLimitInstance(self, logmethod, msg)

    def __enter__(self):
        self._enter()

    def _enter(self):
        with self.lock:
            total_delay = 0.0
            if self.last_ts is None:
                self.last_ts = time.monotonic()
                return total_delay
            while True:
                now = time.monotonic()
                delta = now - self.last_ts
                if delta >= self.delta:
                    break
                delay = self.delta - delta
                time.sleep(delay)
                total_delay += delay
            self.last_ts = time.monotonic()
            return total_delay

    def __exit__(self, etype, value, tb):
        self._exit(etype, value, tb)

    def _exit(self, etype, value, tb):
        pass


class LazyExecutorTTLCache:
    """This is a lazy executor TTL cache.

    It's lazy because if it has cached data, it will always return it
    instantly.

    It's executor based, which means that if a cache miss occurs, it
    will submit a task to an executor to fetch new data.

    Finally, it's a TTL cache, which means it automatically expires data.

    Since it is only expected to be used when caching provider
    resource listing methods, it assumes there will only be one entry
    and ignores arguments -- it will return the same cached data no
    matter what arguments are supplied; but it will pass on those
    arguments to the underlying method in a cache miss.

    :param numeric ttl: The cache timeout in seconds.
    :param concurrent.futures.Executor executor: An executor to use to
        update data asynchronously in case of a cache miss.
    """

    def __init__(self, ttl, executor):
        self.ttl = ttl
        self.executor = executor
        # A lock to make all of this thread safe (especially to ensure
        # we don't fire off multiple updates).
        self.lock = threading.Lock()
        self.cache_clear()

    def cache_clear(self):
        with self.lock:
            # If we have an outstanding update being run by the executor,
            # this is the future.
            self.future = None
            # The last time the underlying method completed.
            self.last_time = None
            # The last value from the underlying method.
            self.last_value = None

    def __call__(self, func):
        def decorator(*args, **kw):
            with self.lock:
                now = time.monotonic()
                if self.future and self.future.done():
                    # If a previous call spawned an update, resolve
                    # that now so we can use the data.
                    try:
                        self.last_time, self.last_value = self.future.result()
                    finally:
                        # Clear the future regardless so we don't loop.
                        self.future = None
                if (self.last_time is not None and
                    now - self.last_time < self.ttl):
                    # A cache hit.
                    return self.last_value
                # The rest of the method is a cache miss.
                if self.last_time is not None:
                    if not self.future:
                        # Fire off an asynchronous update request.
                        # This second wrapper ensures that we record
                        # the time that the update is complete along
                        # with the value.
                        def func_with_time():
                            ret = func(*args, **kw)
                            now = time.monotonic()
                            return (now, ret)
                        self.future = self.executor.submit(func_with_time)
                else:
                    # This is the first time this method has been
                    # called; since we don't have any cached data, we
                    # will synchronously update the data.
                    self.last_value = func(*args, **kw)
                    self.last_time = time.monotonic()
                return self.last_value
        decorator.cache_clear = self.cache_clear
        return decorator


class Segment:
    def __init__(self, index, offset, data):
        self.index = index
        self.offset = offset
        self.data = data


class ImageUploader:
    """
    A helper class for drivers that upload large images in chunks.
    """

    # These values probably don't need to be changed
    error_retries = 3
    concurrency = 10

    # Subclasses must implement these
    segment_size = None

    def __init__(self, endpoint, log, path, image_name, metadata):
        if self.segment_size is None:
            raise Exception("Subclass must set block size")
        self.endpoint = endpoint
        self.log = log
        self.path = path
        self.size = os.path.getsize(path)
        self.image_name = image_name
        self.metadata = metadata
        self.timeout = None

    def shouldRetryException(self, exception):
        return True

    def uploadSegment(self, segment):
        pass

    def startUpload(self):
        pass

    def finishUpload(self):
        pass

    def abortUpload(self):
        pass

    # Main API
    def upload(self, timeout=None):
        if timeout:
            self.timeout = time.monotonic() + timeout
        self.startUpload()
        try:
            with concurrent.futures.ThreadPoolExecutor(
                    max_workers=self.concurrency) as executor:
                with open(self.path, 'rb') as image_file:
                    self._uploadInner(executor, image_file)
            return self.finishUpload()
        except Exception:
            self.log.exception("Error uploading image:")
            self.abortUpload()

    # Subclasses can use this helper method for wrapping retryable calls
    def retry(self, func, *args, **kw):
        for x in range(self.error_retries):
            try:
                return func(*args, **kw)
            except Exception as e:
                if not self.shouldRetryException(e):
                    raise
                if x + 1 >= self.error_retries:
                    raise
                time.sleep(2 * x)

    def getTimeout(self):
        if self.timeout is None:
            return None
        return self.timeout - time.monotonic()

    def checkTimeout(self):
        if self.timeout is None:
            return
        if self.getTimeout() < 0:
            raise Exception("Timed out uploading image")

    # Internal methods
    def _uploadInner(self, executor, image_file):
        futures = set()
        for index, offset in enumerate(range(0, self.size, self.segment_size)):
            segment = Segment(index, offset,
                              image_file.read(self.segment_size))
            future = executor.submit(self.uploadSegment, segment)
            futures.add(future)
            # Keep the pool of workers supplied with data but without
            # reading the entire file into memory.
            if len(futures) >= (self.concurrency * 2):
                (done, futures) = concurrent.futures.wait(
                    futures,
                    timeout=self.getTimeout(),
                    return_when=concurrent.futures.FIRST_COMPLETED)
                for future in done:
                    future.result()
                # Only check the timeout after waiting (not every pass
                # through the loop)
                self.checkTimeout()
        # We're done reading the file, wait for all uploads to finish
        (done, futures) = concurrent.futures.wait(
            futures,
            timeout=self.getTimeout())
        for future in done:
            future.result()
        self.checkTimeout()


class Timer:
    def __init__(self, log, msg):
        self.log = log
        self.msg = msg

    def __enter__(self):
        self.start = time.perf_counter()

    def __exit__(self, type, value, traceback):
        delta = time.perf_counter() - self.start
        self.log.debug(f'{self.msg} in {delta}')



# Copyright 2011 OpenStack, LLC.
# Copyright 2012 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os
import git
import time
import logging
import urllib

from zuul.connection import BaseConnection, ZKChangeCacheMixin
from zuul.driver.git.gitmodel import GitTriggerEvent
from zuul.driver.git.gitwatcher import GitWatcher
from zuul.model import Ref, Branch
from zuul.zk.change_cache import (
    AbstractChangeCache,
    ConcurrentUpdateError,
)


class GitChangeCache(AbstractChangeCache):
    log = logging.getLogger("zuul.driver.GitChangeCache")

    CHANGE_TYPE_MAP = {
        "Ref": Ref,
        "Branch": Branch,
    }


class GitConnection(ZKChangeCacheMixin, BaseConnection):
    driver_name = 'git'
    log = logging.getLogger("zuul.connection.git")

    def __init__(self, driver, connection_name, connection_config):
        super(GitConnection, self).__init__(driver, connection_name,
                                            connection_config)
        if 'baseurl' not in self.connection_config:
            raise Exception('baseurl is required for git connections in '
                            '%s' % self.connection_name)
        self.watcher_thread = None
        self.baseurl = self.connection_config.get('baseurl')
        self.poll_timeout = float(
            self.connection_config.get('poll_delay', 3600 * 2))
        self.canonical_hostname = self.connection_config.get(
            'canonical_hostname')
        if not self.canonical_hostname:
            r = urllib.parse.urlparse(self.baseurl)
            if r.hostname:
                self.canonical_hostname = r.hostname
            else:
                self.canonical_hostname = 'localhost'
        self.projects = {}
        self.source = driver.getSource(self)

    def toDict(self):
        d = super().toDict()
        d.update({
            "baseurl": self.baseurl,
            "canonical_hostname": self.canonical_hostname,
        })
        return d

    def getProject(self, name):
        return self.projects.get(name)

    def addProject(self, project):
        self.projects[project.name] = project

    def getChangeFilesUpdated(self, project_name, branch, tosha):
        job = self.sched.merger.getFilesChangesRaw(
            self.connection_name, project_name, branch, tosha,
            needs_result=True)
        self.log.debug("Waiting for fileschanges job %s" % job)
        job.wait()
        if not job.updated:
            raise Exception("Fileschanges job %s failed" % job)
        self.log.debug("Fileschanges job %s got changes on files %s" %
                       (job, job.files[0]))
        return job.files[0]

    def lsRemote(self, project):
        refs = {}
        client = git.cmd.Git()
        output = client.ls_remote(
            "--heads", "--tags",
            os.path.join(self.baseurl, project))
        for line in output.splitlines():
            sha, ref = line.split('\t')
            if ref.startswith('refs/'):
                refs[ref] = sha
        return refs

    def getChange(self, change_key, refresh=False, event=None):
        change = self._change_cache.get(change_key)
        if change:
            return change

        project = self.source.getProject(change_key.project_name)
        if change_key.change_type == 'Branch':
            branch = change_key.stable_id
            change = Branch(project)
            change.branch = branch
            change.ref = f'refs/heads/{branch}'
            change.oldrev = change_key.oldrev
            change.newrev = change_key.newrev
            change.url = ""
            change.files = self.getChangeFilesUpdated(
                change_key.project_name, branch, change_key.oldrev)
        elif change_key.change_type == 'Ref':
            # catch-all ref (ie, not a branch or head)
            change = Ref(project)
            change.ref = change_key.stable_id
            change.oldrev = change_key.oldrev
            change.newrev = change_key.newrev
            change.url = ""
        else:
            self.log.warning("Unable to get change for %s", change_key)
            return None

        try:
            self._change_cache.set(change_key, change)
        except ConcurrentUpdateError:
            change = self._change_cache.get(change_key)
        return change

    def getProjectBranches(self, project, tenant, min_ltime=-1):
        refs = self.lsRemote(project.name)
        branches = [ref[len('refs/heads/'):] for ref in
                    refs if ref.startswith('refs/heads/')]
        return branches

    def getRefSha(self, project, ref):
        refs = self.lsRemote(project.name)
        return refs.get(ref, '')

    def getGitUrl(self, project):
        return os.path.join(self.baseurl, project.name)

    def watcherCallback(self, data):
        event = GitTriggerEvent()
        event.connection_name = self.connection_name
        event.type = 'ref-updated'
        event.timestamp = time.time()
        event.project_hostname = self.canonical_hostname
        event.project_name = data['project']
        for attr in ('ref', 'oldrev', 'newrev', 'branch_created',
                     'branch_deleted', 'branch_updated'):
            if attr in data:
                setattr(event, attr, data[attr])

        # Force changes cache update before passing
        # the event to the scheduler
        change_key = self.source.getChangeKey(event)
        self.getChange(change_key)
        self.logEvent(event)
        # Pass the event to the scheduler
        self.sched.addTriggerEvent(self.driver_name, event)

    def onLoad(self, zk_client, component_registry):
        self.log.debug("Creating Zookeeper change cache")
        self._change_cache = GitChangeCache(zk_client, self)

        # If the connection was not loaded by a scheduler, but by e.g.
        # zuul-web, we want to stop here.
        if not self.sched:
            return

        self.log.debug("Starting Git Watcher")
        self._start_watcher_thread()

    def onStop(self):
        self.log.debug("Stopping Git Watcher")
        self._stop_watcher_thread()
        if self._change_cache:
            self._change_cache.stop()

    def _stop_watcher_thread(self):
        if self.watcher_thread:
            self.watcher_thread.stop()
            self.watcher_thread.join()

    def _start_watcher_thread(self):
        self.watcher_thread = GitWatcher(
            self,
            self.baseurl,
            self.poll_timeout,
            self.watcherCallback)
        self.watcher_thread.start()



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging
import threading
import time

from collections import defaultdict
from zuul import model
from zuul.lib import tracing
from zuul.lib.logutil import get_annotated_logger
from zuul.lib.statsd import normalize_statsd_name
from zuul.zk.event_queues import (
    PipelineResultEventQueue,
    NodepoolEventElection
)
from zuul.zk.exceptions import LockException
from zuul.zk.nodepool import NodeRequestEvent, ZooKeeperNodepool


class Nodepool(object):
    log = logging.getLogger('zuul.nodepool')
    # The kind of resources we report stats on.  We need a complete
    # list in order to report 0 level gauges.
    resource_types = ('ram', 'cores', 'instances')

    def __init__(self, zk_client, system_id, statsd, scheduler=False):
        self._stopped = False
        self.system_id = system_id
        self.statsd = statsd

        self.election_won = False
        if scheduler:
            # Only enable the node request cache/callback for the scheduler.
            self.stop_watcher_event = threading.Event()
            self.zk_nodepool = ZooKeeperNodepool(
                zk_client,
                enable_node_request_cache=True,
                node_request_event_callback=self._handleNodeRequestEvent,
                connection_suspended_callback=self.stop_watcher_event.set,
                enable_node_cache=True)
            self.election = NodepoolEventElection(zk_client)
            self.event_thread = threading.Thread(target=self.runEventElection)
            self.event_thread.daemon = True
            self.event_thread.start()
        else:
            self.stop_watcher_event = None
            self.zk_nodepool = ZooKeeperNodepool(zk_client)
            self.election = None
            self.event_thread = None

        self.pipeline_result_events = PipelineResultEventQueue.createRegistry(
            zk_client
        )

    def addResources(self, target, source):
        for key, value in source.items():
            if key in self.resource_types:
                target[key] += value

    def runEventElection(self):
        while not self._stopped:
            try:
                self.log.debug("Running nodepool watcher election")
                self.election.run(self._electionWon)
            except Exception:
                self.log.exception("Error in nodepool watcher:")

    def stop(self):
        self.log.debug("Stopping")
        self._stopped = True
        if self.election:
            self.election.cancel()
            self.stop_watcher_event.set()
            self.event_thread.join()
            # Delete the election to avoid a FD leak in tests.
            del self.election

    def _sendNodesProvisionedEvent(self, request):
        tracing.endSavedSpan(request.span_info, attributes={
            "request_id": request.id,
            "state": request.state,
        })
        tenant_name = request.tenant_name
        pipeline_name = request.pipeline_name
        event = model.NodesProvisionedEvent(request.id, request.build_set_uuid)
        self.pipeline_result_events[tenant_name][pipeline_name].put(event)

    def _electionWon(self):
        self.log.info("Watching nodepool requests")
        # Iterate over every completed request in case we are starting
        # up or missed something in the transition.
        self.election_won = True
        try:
            for rid in self.zk_nodepool.getNodeRequests():
                request = self.zk_nodepool.getNodeRequest(rid, cached=True)
                if request.requestor != self.system_id:
                    continue
                if (request.state in {model.STATE_FULFILLED,
                                      model.STATE_FAILED}):
                    self._sendNodesProvisionedEvent(request)
            # Now resume normal event processing.
            self.stop_watcher_event.wait()
        finally:
            self.stop_watcher_event.clear()
            self.election_won = False

    def _handleNodeRequestEvent(self, request, event):
        log = get_annotated_logger(self.log, event=request.event_id)

        if request.requestor != self.system_id:
            return

        log.debug("Node request %s %s", request, request.state)
        if event == NodeRequestEvent.COMPLETED:
            try:
                if self.election_won:
                    if self.election.is_still_valid():
                        self.emitStats(request)
                        self._sendNodesProvisionedEvent(request)
                    else:
                        self.stop_watcher_event.set()
            except Exception:
                # If there are any errors moving the event, re-run the
                # election.
                if self.stop_watcher_event is not None:
                    self.stop_watcher_event.set()
                raise

    def emitStats(self, request):
        try:
            return self._emitStats(request)
        except Exception:
            self.log.exception("Failed to emit stats:")

    def _emitStats(self, request):
        # Implements the following :
        #  counter zuul.nodepool.requests.<state>.total
        #  counter zuul.nodepool.requests.<state>.label.<label>
        #  counter zuul.nodepool.requests.<state>.size.<size>
        #  timer   zuul.nodepool.requests.(fulfilled|failed)
        #  timer   zuul.nodepool.requests.(fulfilled|failed).<label>
        #  timer   zuul.nodepool.requests.(fulfilled|failed).<size>
        if not self.statsd:
            return
        pipe = self.statsd.pipeline()
        state = request.state
        dt = None

        if request.canceled:
            state = 'canceled'
        elif request.state in (model.STATE_FULFILLED, model.STATE_FAILED):
            dt = (request.state_time - request.requested_time) * 1000

        key = 'zuul.nodepool.requests.%s' % state
        pipe.incr(key + ".total")

        if dt:
            pipe.timing(key, dt)
        for label in request.labels:
            safe_label = normalize_statsd_name(label)
            pipe.incr(key + '.label.%s' % safe_label)
            if dt:
                pipe.timing(key + '.label.%s' % safe_label, dt)
        pipe.incr(key + '.size.%s' % len(request.labels))
        if dt:
            pipe.timing(key + '.size.%s' % len(request.labels), dt)
        pipe.send()

    def emitStatsResourceCounters(self, tenant, project, resources, duration):
        if not self.statsd:
            return

        for resource, value in resources.items():
            key = 'zuul.nodepool.resources.in_use.tenant.{tenant}.{resource}'
            self.statsd.incr(
                key, value * duration, tenant=tenant, resource=resource)
        for resource, value in resources.items():
            key = 'zuul.nodepool.resources.in_use.project.' \
                  '{project}.{resource}'
            self.statsd.incr(
                key, value * duration, project=project, resource=resource)

    def requestNodes(self, build_set_uuid, job, tenant_name, pipeline_name,
                     provider, priority, relative_priority, event=None):
        log = get_annotated_logger(self.log, event)
        labels = [n.label for n in job.nodeset.getNodes()]
        if event:
            event_id = event.zuul_event_id
        else:
            event_id = None
        req = model.NodeRequest(self.system_id, build_set_uuid, tenant_name,
                                pipeline_name, job.uuid, job.name, labels,
                                provider, relative_priority, event_id)

        if job.nodeset.nodes:
            self.zk_nodepool.submitNodeRequest(req, priority)
            # Logged after submission so that we have the request id
            log.info("Submitted node request %s", req)
            self.emitStats(req)
        else:
            # Directly fulfill the node request without submitting it to ZK,
            # so nodepool doesn't have to deal with it.
            log.info("Fulfilling empty node request %s", req)
            req.state = model.STATE_FULFILLED
        return req

    def cancelRequest(self, request):
        log = get_annotated_logger(self.log, request.event_id)
        log.info("Canceling node request %s", request)
        try:
            request.canceled = True
            if self.zk_nodepool.deleteNodeRequest(request.id):
                self.emitStats(request)
        except Exception:
            log.exception("Error deleting node request:")

    def reviseRequest(self, request, relative_priority=None):
        '''Attempt to update the node request, if it is not currently being
        processed.

        :param: NodeRequest request: The request to update.
        :param relative_priority int: If supplied, the new relative
            priority to set on the request.

        '''
        log = get_annotated_logger(self.log, request.event_id)
        if relative_priority is None:
            return
        try:
            self.zk_nodepool.lockNodeRequest(request, blocking=False)
        except LockException:
            # It may be locked by nodepool, which is fine.
            log.debug("Unable to revise locked node request %s", request)
            return False
        try:
            old_priority = request.relative_priority
            request.relative_priority = relative_priority
            self.zk_nodepool.storeNodeRequest(request)
            log.debug("Revised relative priority of "
                      "node request %s from %s to %s",
                      request, old_priority, relative_priority)
        except Exception:
            log.exception("Unable to update node request %s", request)
        finally:
            try:
                self.zk_nodepool.unlockNodeRequest(request)
            except Exception:
                log.exception("Unable to unlock node request %s", request)

    def isNodeRequestID(self, request_id):
        try:
            # Node requests are of the form '100-0000000001' with the third
            # character separating the request priority from the sequence
            # number.
            return request_id[3] == "-"
        except Exception:
            return False

    def holdNodeSet(self, nodeset, request, build, duration,
                    zuul_event_id=None):
        '''
        Perform a hold on the given set of nodes.

        :param NodeSet nodeset: The object containing the set of nodes to hold.
        :param HoldRequest request: Hold request associated with the NodeSet
        '''
        log = get_annotated_logger(self.log, zuul_event_id)
        log.info("Holding nodeset %s", nodeset)
        resources = defaultdict(int)
        nodes = nodeset.getNodes()

        log.info(
            "Nodeset %s with %s nodes was in use for %s seconds for build %s "
            "for project %s",
            nodeset, len(nodeset.nodes), duration, build, request.project)

        for node in nodes:
            if node.lock is None:
                raise Exception(f"Node {node} is not locked")
            if node.resources:
                self.addResources(resources, node.resources)
            node.state = model.STATE_HOLD
            node.hold_job = " ".join([request.tenant,
                                      request.project,
                                      request.job,
                                      request.ref_filter])
            node.comment = request.reason
            if request.node_expiration:
                node.hold_expiration = request.node_expiration
            self.zk_nodepool.storeNode(node)

        request.nodes.append(dict(
            build=build.uuid,
            nodes=[node.id for node in nodes],
        ))
        request.current_count += 1

        # Request has been used at least the maximum number of times so set
        # the expiration time so that it can be auto-deleted.
        if request.current_count >= request.max_count and not request.expired:
            request.expired = time.time()

        # Give ourselves a few seconds to try to obtain the lock rather than
        # immediately give up.
        self.zk_nodepool.lockHoldRequest(request, timeout=5)

        try:
            self.zk_nodepool.storeHoldRequest(request)
        except Exception:
            # If we fail to update the request count, we won't consider it
            # a real autohold error by passing the exception up. It will
            # just get used more than the original count specified.
            # It's possible to leak some held nodes, though, which would
            # require manual node deletes.
            log.exception("Unable to update hold request %s:", request)
        finally:
            # Although any exceptions thrown here are handled higher up in
            # _doBuildCompletedEvent, we always want to try to unlock it.
            self.zk_nodepool.unlockHoldRequest(request)

        if resources and duration:
            self.emitStatsResourceCounters(
                request.tenant, request.project, resources, duration)

    def useNodeSet(self, nodeset, tenant_name, project_name,
                   zuul_event_id=None):
        log = get_annotated_logger(self.log, zuul_event_id)
        log.info("Setting nodeset %s in use", nodeset)
        user_data = dict(
            zuul_system=self.system_id,
            project_name=project_name,
        )
        for node in nodeset.getNodes():
            if node.lock is None:
                raise Exception("Node %s is not locked", node)
            node.state = model.STATE_IN_USE
            node.user_data = user_data
            self.zk_nodepool.storeNode(node)

    def returnNodeSet(self, nodeset, build, tenant_name, project_name,
                      duration, zuul_event_id=None):
        log = get_annotated_logger(self.log, zuul_event_id)
        log.info("Returning nodeset %s", nodeset)
        resources = defaultdict(int)

        for node in nodeset.getNodes():
            if node.lock is None:
                log.error("Node %s is not locked", node)
            else:
                try:
                    if node.state == model.STATE_IN_USE:
                        if node.resources:
                            self.addResources(resources, node.resources)
                        node.state = model.STATE_USED
                        self.zk_nodepool.storeNode(node)
                except Exception:
                    log.exception("Exception storing node %s "
                                  "while unlocking:", node)
        self.unlockNodeSet(nodeset)

        log.info("Nodeset %s with %s nodes was in use "
                 "for %s seconds for build %s for project %s",
                 nodeset, len(nodeset.nodes), duration, build, project_name)

        if resources and duration:
            self.emitStatsResourceCounters(
                tenant_name, project_name, resources, duration)

    def unlockNodeSet(self, nodeset):
        self._unlockNodes(nodeset.getNodes())

    def _unlockNodes(self, nodes):
        for node in nodes:
            try:
                self.zk_nodepool.unlockNode(node)
            except Exception:
                self.log.exception("Error unlocking node:")

    def lockNodes(self, request, nodeset):
        log = get_annotated_logger(self.log, event=request.event_id)
        # Try to lock all of the supplied nodes.  If any lock fails,
        # try to unlock any which have already been locked before
        # re-raising the error.
        locked_nodes = []
        try:
            for node_id, node in zip(request.nodes, nodeset.getNodes()):
                self.zk_nodepool.updateNode(node, node_id)
                if node.allocated_to != request.id:
                    raise Exception("Node %s allocated to %s, not %s" %
                                    (node.id, node.allocated_to, request.id))
                log.debug("Locking node %s", node)
                self.zk_nodepool.lockNode(node, timeout=30)
                # Check the allocated_to again to ensure that nodepool didn't
                # re-allocate the nodes to a different node request while we
                # were locking them.
                if node.allocated_to != request.id:
                    raise Exception(
                        "Node %s was reallocated during locking %s, not %s" %
                        (node.id, node.allocated_to, request.id))
                locked_nodes.append(node)
        except Exception:
            log.exception("Error locking nodes:")
            self._unlockNodes(locked_nodes)
            raise

    def getNodesetInfo(self, request, job_nodeset):
        """
        Get a NodesetInfo object with data about real nodes.

        :param NodeRequest request: The fulfilled NodeRequest
        :param NodeSet job_nodeset: The NodeSet object attached to the job

        :returns: A new NodesetInfo object which contains information from
            nodepool about the actual allocated nodes (if any).
        """
        # A copy of the nodeset with information about the real nodes
        nodeset = job_nodeset.copy()

        # If we didn't request nodes and the request is fulfilled then just
        # return. We don't have to do anything in this case.
        if not request.labels and request.fulfilled:
            return model.NodesetInfo()

        # Load the node info from ZK.
        for node_id, node in zip(request.nodes, nodeset.getNodes()):
            self.zk_nodepool.updateNode(node, node_id)

        info = {}
        node = nodeset.getNodes()[0]
        if node.attributes:
            info['zone'] = node.attributes.get('executor-zone')
        else:
            info['zone'] = None
        info['provider'] = node.provider
        info['nodes'] = [n.id for n in nodeset.getNodes()]

        return model.NodesetInfo(**info)

    def acceptNodes(self, request, nodeset):
        # Accept the nodes supplied by request, mutate nodeset with
        # the real node information.
        locked = False
        if request.fulfilled:
            # If the request succeeded, try to lock the nodes.
            try:
                nodes = self.lockNodes(request, nodeset)
                locked = True
            except Exception:
                log = get_annotated_logger(self.log, request.event_id)
                log.exception("Error locking nodes:")
                request.failed = True

        # Regardless of whether locking (or even the request)
        # succeeded, delete the request.
        if not self.deleteNodeRequest(request.id, locked, request.event_id):
            request.failed = True
            self.unlockNodeSet(request.nodeset)

        if request.failed:
            raise Exception("Accepting nodes failed")
        return nodes

    def deleteNodeRequest(self, request_id, locked=False, event_id=None):
        log = get_annotated_logger(self.log, event_id)
        log.debug("Deleting node request %s", request_id)
        try:
            self.zk_nodepool.deleteNodeRequest(request_id)
        except Exception:
            log.exception("Error deleting node request:")
            return False
        return True

    def getNodeRequests(self):
        """Get all node requests submitted by Zuul

        Note this relies entirely on the internal cache.

        :returns: An iterator of NodeRequest objects created by this Zuul
            system.
        """
        for req_id in self.zk_nodepool.getNodeRequests(cached=True):
            req = self.zk_nodepool.getNodeRequest(req_id, cached=True)
            if req.requestor == self.system_id:
                yield req

    def getNodes(self):
        """Get all nodes in use or held by Zuul

        Note this relies entirely on the internal cache.

        :returns: An iterator of Node objects in use (or held) by this Zuul
            system.
        """
        for node_id in self.zk_nodepool.getNodes(cached=True):
            node = self.zk_nodepool.getNode(node_id)
            if (node.user_data and
                isinstance(node.user_data, dict) and
                node.user_data.get('zuul_system') == self.system_id):
                yield node

    def emitStatsTotals(self, abide):
        if not self.statsd:
            return

        total_requests = 0
        tenant_requests = defaultdict(int)
        in_use_resources_by_project = {}
        in_use_resources_by_tenant = {}
        total_resources_by_tenant = {}
        empty_resource_dict = dict([(k, 0) for k in self.resource_types])

        # Initialize zero values for gauges
        for tenant in abide.tenants.values():
            tenant_requests[tenant.name] = 0
            in_use_resources_by_tenant[tenant.name] =\
                empty_resource_dict.copy()
            total_resources_by_tenant[tenant.name] = empty_resource_dict.copy()
            for project in tenant.all_projects:
                in_use_resources_by_project[project.canonical_name] =\
                    empty_resource_dict.copy()

        # Count node requests
        for req in self.getNodeRequests():
            total_requests += 1
            if not req.tenant_name:
                continue
            tenant_requests[req.tenant_name] += 1

        self.statsd.gauge('zuul.nodepool.current_requests',
                          total_requests)
        for tenant, request_count in tenant_requests.items():
            self.statsd.gauge(
                "zuul.nodepool.tenant.{tenant}.current_requests",
                request_count,
                tenant=tenant)

        # Count nodes
        for node in self.zk_nodepool.nodeIterator(cached=True):
            if not node.resources:
                continue

            tenant_name = node.tenant_name
            if tenant_name in total_resources_by_tenant and \
               node.requestor == self.system_id:
                self.addResources(
                    total_resources_by_tenant[tenant_name],
                    node.resources)

            # below here, we are only interested in nodes which are either
            # in-use, used, or currently held by this zuul system
            if node.state not in {model.STATE_IN_USE,
                                  model.STATE_USED,
                                  model.STATE_HOLD}:
                continue
            if not node.user_data:
                continue
            if not isinstance(node.user_data, dict):
                continue
            if node.user_data.get('zuul_system') != self.system_id:
                continue

            if tenant_name in in_use_resources_by_tenant:
                self.addResources(
                    in_use_resources_by_tenant[tenant_name],
                    node.resources)

            project_name = node.user_data.get('project_name')
            if project_name in in_use_resources_by_project:
                self.addResources(
                    in_use_resources_by_project[project_name],
                    node.resources)

        for tenant, resources in total_resources_by_tenant.items():
            for resource, value in resources.items():
                key = 'zuul.nodepool.resources.total.tenant.' \
                      '{tenant}.{resource}'
                self.statsd.gauge(key, value, tenant=tenant, resource=resource)
        for tenant, resources in in_use_resources_by_tenant.items():
            for resource, value in resources.items():
                key = 'zuul.nodepool.resources.in_use.tenant.' \
                      '{tenant}.{resource}'
                self.statsd.gauge(key, value, tenant=tenant, resource=resource)
        for project, resources in in_use_resources_by_project.items():
            for resource, value in resources.items():
                key = 'zuul.nodepool.resources.in_use.project.' \
                      '{project}.{resource}'
                self.statsd.gauge(
                    key, value, project=project, resource=resource)



# Copyright 2012 Hewlett-Packard Development Company, L.P.
# Copyright 2018 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from tests.base import ZuulTestCase, skipIfMultiScheduler


class TestGerritToGithubCRD(ZuulTestCase):
    config_file = 'zuul-gerrit-github.conf'
    tenant_config_file = 'config/cross-source/main.yaml'

    def test_crd_gate(self):
        "Test cross-repo dependencies"
        A = self.fake_gerrit.addFakeChange('gerrit/project1', 'master', 'A')
        B = self.fake_github.openFakePullRequest('github/project2', 'master',
                                                 'B')

        A.addApproval('Code-Review', 2)

        AM2 = self.fake_gerrit.addFakeChange('gerrit/project1', 'master',
                                             'AM2')
        AM1 = self.fake_gerrit.addFakeChange('gerrit/project1', 'master',
                                             'AM1')
        AM2.setMerged()
        AM1.setMerged()

        # A -> AM1 -> AM2
        # A Depends-On: B
        # M2 is here to make sure it is never queried.  If it is, it
        # means zuul is walking down the entire history of merged
        # changes.

        A.setDependsOn(AM1, 1)
        AM1.setDependsOn(AM2, 1)

        A.data['commitMessage'] = '%s\n\nDepends-On: %s\n' % (
            A.subject, B.url)

        self.fake_gerrit.addEvent(A.addApproval('Approved', 1))
        self.waitUntilSettled()

        self.assertEqual(A.data['status'], 'NEW')
        self.assertFalse(B.is_merged)

        for connection in self.scheds.first.connections.connections.values():
            connection.maintainCache([], max_age=0)

        self.executor_server.hold_jobs_in_build = True
        B.addLabel('approved')
        self.fake_gerrit.addEvent(A.addApproval('Approved', 1))
        self.waitUntilSettled()

        self.executor_server.release('.*-merge')
        self.waitUntilSettled()
        self.executor_server.release('.*-merge')
        self.waitUntilSettled()
        self.executor_server.hold_jobs_in_build = False
        self.executor_server.release()
        self.waitUntilSettled()

        self.assertEqual(AM2.queried, 0)
        self.assertEqual(A.data['status'], 'MERGED')
        self.assertTrue(B.is_merged)
        self.assertEqual(A.reported, 2)
        self.assertEqual(len(B.comments), 2)

        changes = self.getJobFromHistory(
            'project-merge', 'gerrit/project1').changes
        self.assertEqual(changes, '1,%s 1,1' % B.head_sha)

    def test_crd_branch(self):
        "Test cross-repo dependencies in multiple branches"

        self.create_branch('github/project2', 'mp')
        A = self.fake_gerrit.addFakeChange('gerrit/project1', 'master', 'A')
        B = self.fake_github.openFakePullRequest('github/project2', 'master',
                                                 'B')
        C1 = self.fake_github.openFakePullRequest('github/project2', 'mp',
                                                  'C1')

        A.addApproval('Code-Review', 2)

        # A Depends-On: B+C1
        A.data['commitMessage'] = '%s\n\nDepends-On: %s\nDepends-On: %s\n' % (
            A.subject, B.url, C1.url)

        self.executor_server.hold_jobs_in_build = True
        B.addLabel('approved')
        C1.addLabel('approved')
        self.fake_gerrit.addEvent(A.addApproval('Approved', 1))
        self.waitUntilSettled()

        self.executor_server.release('.*-merge')
        self.waitUntilSettled()
        self.executor_server.release('.*-merge')
        self.waitUntilSettled()
        self.executor_server.release('.*-merge')
        self.waitUntilSettled()
        self.executor_server.hold_jobs_in_build = False
        self.executor_server.release()
        self.waitUntilSettled()

        self.assertEqual(A.data['status'], 'MERGED')
        self.assertTrue(B.is_merged)
        self.assertTrue(C1.is_merged)
        self.assertEqual(A.reported, 2)
        self.assertEqual(len(B.comments), 2)
        self.assertEqual(len(C1.comments), 2)

        changes = self.getJobFromHistory(
            'project-merge', 'gerrit/project1').changes
        self.assertEqual(changes, '1,%s 2,%s 1,1' %
                         (B.head_sha, C1.head_sha))

    def test_crd_gate_reverse(self):
        "Test reverse cross-repo dependencies"
        A = self.fake_gerrit.addFakeChange('gerrit/project1', 'master', 'A')
        B = self.fake_github.openFakePullRequest('github/project2', 'master',
                                                 'B')
        A.addApproval('Code-Review', 2)

        # A Depends-On: B

        A.data['commitMessage'] = '%s\n\nDepends-On: %s\n' % (
            A.subject, B.url)

        self.fake_gerrit.addEvent(A.addApproval('Approved', 1))
        self.waitUntilSettled()

        self.assertEqual(A.data['status'], 'NEW')
        self.assertFalse(B.is_merged)

        self.executor_server.hold_jobs_in_build = True
        A.addApproval('Approved', 1)
        self.fake_github.emitEvent(B.addLabel('approved'))
        self.waitUntilSettled()

        self.executor_server.release('.*-merge')
        self.waitUntilSettled()
        self.executor_server.release('.*-merge')
        self.waitUntilSettled()
        self.executor_server.hold_jobs_in_build = False
        self.executor_server.release()
        self.waitUntilSettled()

        self.assertEqual(A.data['status'], 'MERGED')
        self.assertTrue(B.is_merged)
        self.assertEqual(A.reported, 2)
        self.assertEqual(len(B.comments), 2)

        changes = self.getJobFromHistory(
            'project-merge', 'gerrit/project1').changes
        self.assertEqual(changes, '1,%s 1,1' %
                         (B.head_sha,))

    def test_crd_cycle(self):
        "Test cross-repo dependency cycles"
        A = self.fake_gerrit.addFakeChange('gerrit/project1', 'master', 'A')
        msg = "Depends-On: %s" % (A.data['url'],)
        B = self.fake_github.openFakePullRequest('github/project2', 'master',
                                                 'B', body=msg)
        A.addApproval('Code-Review', 2)
        B.addLabel('approved')

        # A -> B -> A (via commit-depends)

        A.data['commitMessage'] = '%s\n\nDepends-On: %s\n' % (
            A.subject, B.url)

        self.fake_gerrit.addEvent(A.addApproval('Approved', 1))
        self.waitUntilSettled()

        self.assertEqual(A.reported, 1)
        self.assertEqual(
            A.messages[0],
            "Build failed.\n\n\nWarning:\n  "
            "Dependency cycle detected and project github/project2 "
            "doesn't allow circular dependencies\n")
        self.assertEqual(len(B.comments), 0)
        self.assertEqual(A.data['status'], 'NEW')
        self.assertFalse(B.is_merged)

    def test_crd_gate_unknown(self):
        "Test unknown projects in dependent pipeline"
        self.init_repo("github/unknown", tag='init')
        # We also need make the repo known to the fake github as a
        # side-effect of getting the project.
        self.fake_github.source.getProject("github/unknown")
        A = self.fake_gerrit.addFakeChange('gerrit/project1', 'master', 'A')
        B = self.fake_github.openFakePullRequest('github/unknown', 'master',
                                                 'B')
        A.addApproval('Code-Review', 2)

        # A Depends-On: B
        A.data['commitMessage'] = '%s\n\nDepends-On: %s\n' % (
            A.subject, B.url)

        event = B.addLabel('approved')
        self.fake_gerrit.addEvent(A.addApproval('Approved', 1))
        self.waitUntilSettled()

        # Unknown projects cannot share a queue with any other since
        # they don't have common jobs with any other (they have no
        # jobs).  Changes which depend on unknown project changes
        # should only report an error.
        self.assertEqual(A.data['status'], 'NEW')
        self.assertFalse(B.is_merged)
        self.assertEqual(A.reported, 1)
        self.assertIn("does not share a change queue", A.messages[-1])
        self.assertEqual(len(B.comments), 0)
        self.assertEqual(len(self.history), 0)

        # Simulate change B being gated outside this layout Set the
        # change merged before submitting the event so that when the
        # event triggers a gerrit query to update the change, we get
        # the information that it was merged.
        B.setMerged('merged')
        self.fake_github.emitEvent(event)
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 0)

        # Now that B is merged, A should be able to be enqueued and
        # merged.
        self.fake_gerrit.addEvent(A.addApproval('Approved', 1))
        self.waitUntilSettled()

        self.assertEqual(A.data['status'], 'MERGED')
        self.assertEqual(A.reported, 3)
        self.assertTrue(B.is_merged)
        self.assertEqual(len(B.comments), 0)

    def test_crd_check(self):
        "Test cross-repo dependencies in independent pipelines"
        self.executor_server.hold_jobs_in_build = True
        self.hold_jobs_in_queue = True
        A = self.fake_gerrit.addFakeChange('gerrit/project1', 'master', 'A')
        B = self.fake_github.openFakePullRequest(
            'github/project2', 'master', 'B')

        # A Depends-On: B
        A.data['commitMessage'] = '%s\n\nDepends-On: %s\n' % (
            A.subject, B.url)

        self.fake_gerrit.addEvent(A.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()

        self.hold_jobs_in_queue = False
        self.executor_api.release()
        self.waitUntilSettled()

        self.executor_server.release('.*-merge')
        self.waitUntilSettled()

        self.assertTrue(self.builds[0].hasChanges(A, B))

        self.executor_server.hold_jobs_in_build = False
        self.executor_server.release()
        self.waitUntilSettled()

        self.assertEqual(A.data['status'], 'NEW')
        self.assertFalse(B.is_merged)
        self.assertEqual(A.reported, 1)
        self.assertEqual(len(B.comments), 0)

        changes = self.getJobFromHistory(
            'project-merge', 'gerrit/project1').changes
        self.assertEqual(changes, '1,%s 1,1' %
                         (B.head_sha,))

        tenant = self.scheds.first.sched.abide.tenants.get('tenant-one')
        self.assertEqual(
            len(tenant.layout.pipeline_managers['check'].state.queues), 0)

    def test_crd_check_duplicate(self):
        "Test duplicate check in independent pipelines"
        self.executor_server.hold_jobs_in_build = True
        A = self.fake_gerrit.addFakeChange('gerrit/project1', 'master', 'A')
        B = self.fake_github.openFakePullRequest(
            'github/project2', 'master', 'B')
        self.waitUntilSettled()

        # A Depends-On: B
        A.data['commitMessage'] = '%s\n\nDepends-On: %s\n' % (
            A.subject, B.url)

        # Add two dependent changes...
        self.fake_gerrit.addEvent(A.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()
        self.assertEqual(len(self.getAllItems('tenant-one', 'check')), 2)

        # ...make sure the live one is not duplicated...
        self.fake_gerrit.addEvent(A.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()
        self.assertEqual(len(self.getAllItems('tenant-one', 'check')), 2)

        # ...but the non-live one is able to be.
        self.fake_github.emitEvent(B.getPullRequestEditedEvent())
        self.waitUntilSettled()
        self.assertEqual(len(self.getAllItems('tenant-one', 'check')), 3)

        # Release jobs in order to avoid races with change A jobs
        # finishing before change B jobs.
        self.orderedRelease()
        self.executor_server.hold_jobs_in_build = False
        self.executor_server.release()
        self.waitUntilSettled()

        self.assertEqual(A.data['status'], 'NEW')
        self.assertFalse(B.is_merged)
        self.assertEqual(A.reported, 1)
        self.assertEqual(len(B.comments), 1)

        changes = self.getJobFromHistory(
            'project-merge', 'gerrit/project1').changes
        self.assertEqual(changes, '1,%s 1,1' %
                         (B.head_sha,))

        changes = self.getJobFromHistory(
            'project-merge', 'github/project2').changes
        self.assertEqual(changes, '1,%s' %
                         (B.head_sha,))
        self.assertEqual(len(self.getAllQueues('tenant-one', 'check')), 0)

        self.assertIn('Build succeeded', A.messages[0])

    def _test_crd_check_reconfiguration(self, project1, project2):
        "Test cross-repo dependencies re-enqueued in independent pipelines"

        self.hold_jobs_in_queue = True
        A = self.fake_gerrit.addFakeChange(project1, 'master', 'A')
        B = self.fake_github.openFakePullRequest(project2, 'master', 'B')

        # A Depends-On: B
        A.data['commitMessage'] = '%s\n\nDepends-On: %s\n' % (
            A.subject, B.url)

        self.fake_gerrit.addEvent(A.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()

        self.scheds.execute(lambda app: app.sched.reconfigure(app.config))
        self.waitUntilSettled()

        # Make sure the items still share a change queue, and the
        # first one is not live.
        tenant = self.scheds.first.sched.abide.tenants.get('tenant-one')
        self.assertEqual(
            len(tenant.layout.pipeline_managers['check'].state.queues), 1)
        queue = tenant.layout.pipeline_managers['check'].state.queues[0]
        first_item = queue.queue[0]
        for item in queue.queue:
            self.assertEqual(item.queue, first_item.queue)
        self.assertFalse(first_item.live)
        self.assertTrue(queue.queue[1].live)

        self.hold_jobs_in_queue = False
        self.executor_api.release()
        self.waitUntilSettled()

        self.assertEqual(A.data['status'], 'NEW')
        self.assertFalse(B.is_merged)
        self.assertEqual(A.reported, 1)
        self.assertEqual(len(B.comments), 0)

        changes = self.getJobFromHistory(
            'project-merge', project1).changes
        self.assertEqual(changes, '1,%s 1,1' %
                         (B.head_sha,))
        self.assertEqual(
            len(tenant.layout.pipeline_managers['check'].state.queues), 0)

    @skipIfMultiScheduler()
    def test_crd_check_reconfiguration(self):
        self._test_crd_check_reconfiguration(
            'gerrit/project1', 'github/project2')

    @skipIfMultiScheduler()
    def test_crd_undefined_project(self):
        """Test that undefined projects in dependencies are handled for
        independent pipelines"""
        # It's a hack for fake github,
        # as it implies repo creation upon the creation of any change
        self.init_repo("github/unknown", tag='init')
        # We also need make the repo known to the fake github as a
        # side-effect of getting the project.
        self.fake_github.source.getProject("github/unknown")
        self._test_crd_check_reconfiguration('gerrit/project1',
                                             'github/unknown')

    def test_crd_check_transitive(self):
        "Test transitive cross-repo dependencies"
        # Specifically, if A -> B -> C, and C gets a new patchset and
        # A gets a new patchset, ensure the test of A,2 includes B,1
        # and C,2 (not C,1 which would indicate stale data in the
        # cache for B).
        A = self.fake_gerrit.addFakeChange('gerrit/project1', 'master', 'A')
        C = self.fake_gerrit.addFakeChange('gerrit/project1', 'master', 'C')
        # B Depends-On: C
        msg = "Depends-On: %s" % (C.data['url'],)
        B = self.fake_github.openFakePullRequest(
            'github/project2', 'master', 'B', body=msg)

        # A Depends-On: B
        A.data['commitMessage'] = '%s\n\nDepends-On: %s\n' % (
            A.subject, B.url)

        self.fake_gerrit.addEvent(A.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()
        self.assertEqual(self.history[-1].changes, '2,1 1,%s 1,1' %
                         (B.head_sha,))

        self.fake_github.emitEvent(B.getPullRequestEditedEvent())
        self.waitUntilSettled()
        self.assertEqual(self.history[-1].changes, '2,1 1,%s' %
                         (B.head_sha,))

        self.fake_gerrit.addEvent(C.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()
        self.assertEqual(self.history[-1].changes, '2,1')

        C.addPatchset()
        self.fake_gerrit.addEvent(C.getPatchsetCreatedEvent(2))
        self.waitUntilSettled()
        self.assertEqual(self.history[-1].changes, '2,2')

        A.addPatchset()
        self.fake_gerrit.addEvent(A.getPatchsetCreatedEvent(2))
        self.waitUntilSettled()
        self.assertEqual(self.history[-1].changes, '2,2 1,%s 1,2' %
                         (B.head_sha,))

    def test_crd_check_unknown(self):
        "Test unknown projects in independent pipeline"
        self.init_repo("github/unknown", tag='init')
        # We also need make the repo known to the fake github as a
        # side-effect of getting the project.
        self.fake_github.source.getProject("github/unknown")
        A = self.fake_gerrit.addFakeChange('gerrit/project1', 'master', 'A')
        B = self.fake_github.openFakePullRequest(
            'github/unknown', 'master', 'B')
        # A Depends-On: B
        A.data['commitMessage'] = '%s\n\nDepends-On: %s\n' % (
            A.subject, B.url)

        # Make sure zuul has seen an event on B. This is necessary
        # in order to populate our fake github project db.
        self.fake_github.emitEvent(B.getPullRequestEditedEvent())
        # Note we wait until settled here as the event processing for
        # the next event may not have the updated db yet otherwise.
        self.waitUntilSettled()
        self.fake_gerrit.addEvent(A.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()

        self.assertEqual(A.data['status'], 'NEW')
        self.assertEqual(A.reported, 1)
        self.assertFalse(B.is_merged)
        self.assertEqual(len(B.comments), 0)

    def test_crd_cycle_join(self):
        "Test an updated change creates a cycle"
        A = self.fake_github.openFakePullRequest(
            'github/project2', 'master', 'A')

        self.fake_github.emitEvent(A.getPullRequestEditedEvent())
        self.waitUntilSettled()
        self.assertEqual(len(A.comments), 1)

        # Create B->A
        B = self.fake_gerrit.addFakeChange('gerrit/project1', 'master', 'B')
        B.data['commitMessage'] = '%s\n\nDepends-On: %s\n' % (
            B.subject, A.url)
        self.fake_gerrit.addEvent(B.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()

        # Dep is there so zuul should have reported on B
        self.assertEqual(B.reported, 1)

        # Update A to add A->B (a cycle).
        self.fake_github.emitEvent(
            A.editBody('Depends-On: %s\n' % (B.data['url'])))
        self.waitUntilSettled()

        # Dependency cycle injected so zuul should have reported again on A
        self.assertEqual(len(A.comments), 2)

        # Now if we update B to remove the depends-on, everything
        # should be okay.  B; A->B

        B.addPatchset()
        B.data['commitMessage'] = '%s\n' % (B.subject,)
        self.fake_github.emitEvent(A.getPullRequestEditedEvent())
        self.waitUntilSettled()

        # Cycle was removed so now zuul should have reported again on A
        self.assertEqual(len(A.comments), 3)

        self.fake_gerrit.addEvent(B.getPatchsetCreatedEvent(2))
        self.waitUntilSettled()
        self.assertEqual(B.reported, 2)


class TestGithubToGerritCRD(ZuulTestCase):
    config_file = 'zuul-gerrit-github.conf'
    tenant_config_file = 'config/cross-source/main.yaml'

    def test_crd_gate(self):
        "Test cross-repo dependencies"
        A = self.fake_github.openFakePullRequest('github/project2', 'master',
                                                 'A')
        B = self.fake_gerrit.addFakeChange('gerrit/project1', 'master', 'B')

        B.addApproval('Code-Review', 2)

        # A Depends-On: B
        A.editBody('Depends-On: %s\n' % (B.data['url']))

        event = A.addLabel('approved')
        self.fake_github.emitEvent(event)
        self.waitUntilSettled()

        self.assertFalse(A.is_merged)
        self.assertEqual(B.data['status'], 'NEW')

        for connection in self.scheds.first.connections.connections.values():
            connection.maintainCache([], max_age=0)

        self.executor_server.hold_jobs_in_build = True
        B.addApproval('Approved', 1)
        self.fake_github.emitEvent(event)
        self.waitUntilSettled()

        self.executor_server.release('.*-merge')
        self.waitUntilSettled()
        self.executor_server.release('.*-merge')
        self.waitUntilSettled()
        self.executor_server.hold_jobs_in_build = False
        self.executor_server.release()
        self.waitUntilSettled()

        self.assertTrue(A.is_merged)
        self.assertEqual(B.data['status'], 'MERGED')
        self.assertEqual(len(A.comments), 2)
        self.assertEqual(B.reported, 2)

        changes = self.getJobFromHistory(
            'project-merge', 'github/project2').changes
        self.assertEqual(changes, '1,1 1,%s' % A.head_sha)

    def test_crd_branch(self):
        "Test cross-repo dependencies in multiple branches"

        self.create_branch('gerrit/project1', 'mp')
        A = self.fake_github.openFakePullRequest('github/project2', 'master',
                                                 'A')
        B = self.fake_gerrit.addFakeChange('gerrit/project1', 'master', 'B')
        C1 = self.fake_gerrit.addFakeChange('gerrit/project1', 'mp', 'C1')

        B.addApproval('Code-Review', 2)
        C1.addApproval('Code-Review', 2)

        # A Depends-On: B+C1
        A.editBody('Depends-On: %s\nDepends-On: %s\n' % (
            B.data['url'], C1.data['url']))

        self.executor_server.hold_jobs_in_build = True
        B.addApproval('Approved', 1)
        C1.addApproval('Approved', 1)
        self.fake_github.emitEvent(A.addLabel('approved'))
        self.waitUntilSettled()

        self.executor_server.release('.*-merge')
        self.waitUntilSettled()
        self.executor_server.release('.*-merge')
        self.waitUntilSettled()
        self.executor_server.release('.*-merge')
        self.waitUntilSettled()
        self.executor_server.hold_jobs_in_build = False
        self.executor_server.release()
        self.waitUntilSettled()
        self.assertTrue(A.is_merged)
        self.assertEqual(B.data['status'], 'MERGED')
        self.assertEqual(C1.data['status'], 'MERGED')
        self.assertEqual(len(A.comments), 2)
        self.assertEqual(B.reported, 2)
        self.assertEqual(C1.reported, 2)

        changes = self.getJobFromHistory(
            'project-merge', 'github/project2').changes
        self.assertEqual(changes, '1,1 2,1 1,%s' %
                         (A.head_sha,))

    def test_crd_gate_reverse(self):
        "Test reverse cross-repo dependencies"
        A = self.fake_github.openFakePullRequest('github/project2', 'master',
                                                 'A')
        B = self.fake_gerrit.addFakeChange('gerrit/project1', 'master', 'B')
        B.addApproval('Code-Review', 2)

        # A Depends-On: B
        A.editBody('Depends-On: %s\n' % (B.data['url'],))

        self.fake_github.emitEvent(A.addLabel('approved'))
        self.waitUntilSettled()

        self.assertFalse(A.is_merged)
        self.assertEqual(B.data['status'], 'NEW')

        self.executor_server.hold_jobs_in_build = True
        A.addLabel('approved')
        self.fake_gerrit.addEvent(B.addApproval('Approved', 1))
        self.waitUntilSettled()

        self.executor_server.release('.*-merge')
        self.waitUntilSettled()
        self.executor_server.release('.*-merge')
        self.waitUntilSettled()
        self.executor_server.hold_jobs_in_build = False
        self.executor_server.release()
        self.waitUntilSettled()

        self.assertTrue(A.is_merged)
        self.assertEqual(B.data['status'], 'MERGED')
        self.assertEqual(len(A.comments), 2)
        self.assertEqual(B.reported, 2)

        changes = self.getJobFromHistory(
            'project-merge', 'github/project2').changes
        self.assertEqual(changes, '1,1 1,%s' %
                         (A.head_sha,))

    def test_crd_cycle(self):
        "Test cross-repo dependency cycles"
        A = self.fake_github.openFakePullRequest('github/project2', 'master',
                                                 'A')
        B = self.fake_gerrit.addFakeChange('gerrit/project1', 'master', 'B')
        B.data['commitMessage'] = '%s\n\nDepends-On: %s\n' % (
            B.subject, A.url)

        B.addApproval('Code-Review', 2)
        B.addApproval('Approved', 1)

        # A -> B -> A (via commit-depends)
        A.editBody('Depends-On: %s\n' % (B.data['url'],))

        self.fake_github.emitEvent(A.addLabel('approved'))
        self.waitUntilSettled()

        self.assertEqual(len(A.comments), 1)
        self.assertEqual(B.reported, 0)
        self.assertFalse(A.is_merged)
        self.assertEqual(B.data['status'], 'NEW')

    def test_crd_gate_unknown(self):
        "Test unknown projects in dependent pipeline"
        self.init_repo("gerrit/unknown", tag='init')
        A = self.fake_github.openFakePullRequest('github/project2', 'master',
                                                 'A')
        B = self.fake_gerrit.addFakeChange('gerrit/unknown', 'master', 'B')
        B.addApproval('Code-Review', 2)

        # A Depends-On: B
        A.editBody('Depends-On: %s\n' % (B.data['url'],))

        B.addApproval('Approved', 1)
        event = A.addLabel('approved')
        self.fake_github.emitEvent(event)
        self.waitUntilSettled()

        # Unknown projects cannot share a queue with any other
        # since they don't have common jobs with any other (they have no jobs).
        # Changes which depend on unknown project changes
        # should not be processed in dependent pipeline
        self.assertFalse(A.is_merged)
        self.assertEqual(B.data['status'], 'NEW')
        self.assertEqual(len(A.comments), 1)
        self.assertEqual(B.reported, 0)
        self.assertEqual(len(self.history), 0)

        # Simulate change B being gated outside this layout Set the
        # change merged before submitting the event so that when the
        # event triggers a gerrit query to update the change, we get
        # the information that it was merged.
        B.setMerged()
        self.fake_gerrit.addEvent(B.addApproval('Approved', 1))
        self.waitUntilSettled()
        self.assertEqual(len(self.history), 0)

        # Now that B is merged, A should be able to be enqueued and
        # merged.
        self.fake_github.emitEvent(event)
        self.waitUntilSettled()

        self.assertTrue(A.is_merged)
        self.assertEqual(len(A.comments), 3)
        self.assertEqual(B.data['status'], 'MERGED')
        self.assertEqual(B.reported, 0)

    def test_crd_check(self):
        "Test cross-repo dependencies in independent pipelines"
        self.executor_server.hold_jobs_in_build = True
        self.hold_jobs_in_queue = True
        A = self.fake_github.openFakePullRequest('github/project2', 'master',
                                                 'A')
        B = self.fake_gerrit.addFakeChange(
            'gerrit/project1', 'master', 'B')

        # A Depends-On: B
        self.fake_github.emitEvent(
            A.editBody('Depends-On: %s\n' % (B.data['url'],)))
        self.waitUntilSettled()

        self.hold_jobs_in_queue = False
        self.executor_api.release()
        self.waitUntilSettled()

        self.executor_server.release('.*-merge')
        self.waitUntilSettled()

        self.assertTrue(self.builds[0].hasChanges(A, B))

        self.executor_server.hold_jobs_in_build = False
        self.executor_server.release()
        self.waitUntilSettled()

        self.assertFalse(A.is_merged)
        self.assertEqual(B.data['status'], 'NEW')
        self.assertEqual(len(A.comments), 1)
        self.assertEqual(B.reported, 0)

        changes = self.getJobFromHistory(
            'project-merge', 'github/project2').changes
        self.assertEqual(changes, '1,1 1,%s' %
                         (A.head_sha,))

        tenant = self.scheds.first.sched.abide.tenants.get('tenant-one')
        self.assertEqual(
            len(tenant.layout.pipeline_managers['check'].state.queues), 0)

    def test_crd_check_duplicate(self):
        "Test duplicate check in independent pipelines"
        self.executor_server.hold_jobs_in_build = True
        A = self.fake_github.openFakePullRequest('github/project2', 'master',
                                                 'A')
        B = self.fake_gerrit.addFakeChange(
            'gerrit/project1', 'master', 'B')

        # A Depends-On: B
        event = A.editBody('Depends-On: %s\n' % (B.data['url'],))
        tenant = self.scheds.first.sched.abide.tenants.get('tenant-one')

        # Add two dependent changes...
        self.fake_github.emitEvent(event)
        self.waitUntilSettled()
        self.assertEqual(len(self.getAllItems('tenant-one', 'check')), 2)

        # ...make sure the live one is not duplicated...
        self.fake_github.emitEvent(A.getPullRequestEditedEvent())
        self.waitUntilSettled()
        self.assertEqual(len(self.getAllItems('tenant-one', 'check')), 2)

        # ...but the non-live one is able to be.
        self.fake_gerrit.addEvent(B.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()
        self.assertEqual(len(self.getAllItems('tenant-one', 'check')), 3)

        # Release jobs in order to avoid races with change A jobs
        # finishing before change B jobs.
        self.orderedRelease()
        self.executor_server.hold_jobs_in_build = False
        self.executor_server.release()
        self.waitUntilSettled()

        self.assertFalse(A.is_merged)
        self.assertEqual(B.data['status'], 'NEW')
        self.assertEqual(len(A.comments), 1)
        self.assertEqual(B.reported, 1)

        changes = self.getJobFromHistory(
            'project-merge', 'github/project2').changes
        self.assertEqual(changes, '1,1 1,%s' %
                         (A.head_sha,))

        changes = self.getJobFromHistory(
            'project-merge', 'gerrit/project1').changes
        self.assertEqual(changes, '1,1')
        self.assertEqual(
            len(tenant.layout.pipeline_managers['check'].state.queues), 0)

        self.assertIn('Build succeeded', A.comments[0])

    def _test_crd_check_reconfiguration(self, project1, project2):
        "Test cross-repo dependencies re-enqueued in independent pipelines"

        self.hold_jobs_in_queue = True
        A = self.fake_github.openFakePullRequest(project1, 'master', 'A')
        B = self.fake_gerrit.addFakeChange(project2, 'master', 'B')

        # A Depends-On: B
        self.fake_github.emitEvent(
            A.editBody('Depends-On: %s\n' % (B.data['url'],)))
        self.waitUntilSettled()

        self.scheds.execute(lambda app: app.sched.reconfigure(app.config))
        self.waitUntilSettled()

        # Make sure the items still share a change queue, and the
        # first one is not live.
        tenant = self.scheds.first.sched.abide.tenants.get('tenant-one')
        self.assertEqual(
            len(tenant.layout.pipeline_managers['check'].state.queues), 1)
        queue = tenant.layout.pipeline_managers['check'].state.queues[0]
        first_item = queue.queue[0]
        for item in queue.queue:
            self.assertEqual(item.queue, first_item.queue)
        self.assertFalse(first_item.live)
        self.assertTrue(queue.queue[1].live)

        self.hold_jobs_in_queue = False
        self.executor_api.release()
        self.waitUntilSettled()

        self.assertFalse(A.is_merged)
        self.assertEqual(B.data['status'], 'NEW')
        self.assertEqual(len(A.comments), 1)
        self.assertEqual(B.reported, 0)

        changes = self.getJobFromHistory(
            'project-merge', project1).changes
        self.assertEqual(changes, '1,1 1,%s' %
                         (A.head_sha,))
        self.assertEqual(
            len(tenant.layout.pipeline_managers['check'].state.queues), 0)

    @skipIfMultiScheduler()
    def test_crd_check_reconfiguration(self):
        self._test_crd_check_reconfiguration(
            'github/project2', 'gerrit/project1')

    @skipIfMultiScheduler()
    def test_crd_undefined_project(self):
        """Test that undefined projects in dependencies are handled for
        independent pipelines"""
        # It's a hack for fake gerrit,
        # as it implies repo creation upon the creation of any change
        self.init_repo("gerrit/unknown", tag='init')
        self._test_crd_check_reconfiguration('github/project2',
                                             'gerrit/unknown')

    def test_crd_other_tenant_project(self):
        """Test that undefined projects in dependencies are handled for
        independent pipelines"""
        self.hold_jobs_in_queue = True
        A = self.fake_github.openFakePullRequest('github/project2',
                                                 'master', 'A')
        B = self.fake_gerrit.addFakeChange('gerrit/project2', 'master', 'B')

        # A Depends-On: B
        gh_event = A.editBody('Depends-On: %s\n' % (B.data['url'],))
        self.fake_github.emitEvent(gh_event)
        self.waitUntilSettled()

        self.scheds.execute(lambda app: app.sched.reconfigure(app.config))
        self.waitUntilSettled()

        # Make sure the items still share a change queue, and the
        # first one is not live.
        tenant = self.scheds.first.sched.abide.tenants.get('tenant-one')
        self.assertEqual(
            len(tenant.layout.pipeline_managers['check'].state.queues), 1)
        queue = tenant.layout.pipeline_managers['check'].state.queues[0]
        first_item = queue.queue[0]
        for item in queue.queue:
            self.assertEqual(item.queue, first_item.queue)
        self.assertFalse(first_item.live)
        self.assertTrue(queue.queue[1].live)

        self.hold_jobs_in_queue = False
        self.executor_api.release()
        self.waitUntilSettled()

        self.assertFalse(A.is_merged)
        self.assertEqual(B.data['status'], 'NEW')
        self.assertEqual(len(A.comments), 1)
        self.assertEqual(B.reported, 0)

        changes = self.getJobFromHistory(
            'project-merge', 'github/project2').changes
        expected_changes = f'1,1 1,{A.head_sha}'
        self.assertEqual(changes, expected_changes)
        self.assertEqual(
            len(tenant.layout.pipeline_managers['check'].state.queues), 0)

        # Re-check the same PR again
        self.fake_github.emitEvent(gh_event)
        self.waitUntilSettled()

        self.assertHistory([
            dict(name='project-merge', result='SUCCESS',
                 changes=expected_changes),
            dict(name='project-test1', result='SUCCESS',
                 changes=expected_changes),
            dict(name='project-test2', result='SUCCESS',
                 changes=expected_changes),
            dict(name='project1-project2-integration', result='SUCCESS',
                 changes=expected_changes),
            dict(name='project-merge', result='SUCCESS',
                 changes=expected_changes),
            dict(name='project-test1', result='SUCCESS',
                 changes=expected_changes),
            dict(name='project1-project2-integration', result='SUCCESS',
                 changes=expected_changes),
            dict(name='project-test2', result='SUCCESS',
                 changes=expected_changes),
        ], ordered=False)

    def test_crd_check_transitive(self):
        "Test transitive cross-repo dependencies"
        # Specifically, if A -> B -> C, and C gets a new patchset and
        # A gets a new patchset, ensure the test of A,2 includes B,1
        # and C,2 (not C,1 which would indicate stale data in the
        # cache for B).
        A = self.fake_github.openFakePullRequest('github/project2', 'master',
                                                 'A')
        B = self.fake_gerrit.addFakeChange('gerrit/project1', 'master', 'B')
        C = self.fake_github.openFakePullRequest('github/project2', 'master',
                                                 'C')

        # B Depends-On: C
        B.data['commitMessage'] = '%s\n\nDepends-On: %s\n' % (
            B.subject, C.url)

        # A Depends-On: B
        self.fake_github.emitEvent(
            A.editBody('Depends-On: %s\n' % (B.data['url'],)))
        self.waitUntilSettled()
        self.assertEqual(self.history[-1].changes, '2,%s 1,1 1,%s' %
                         (C.head_sha, A.head_sha))

        self.fake_gerrit.addEvent(B.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()
        self.assertEqual(self.history[-1].changes, '2,%s 1,1' %
                         (C.head_sha,))

        self.fake_github.emitEvent(C.getPullRequestEditedEvent())
        self.waitUntilSettled()
        self.assertEqual(self.history[-1].changes, '2,%s' %
                         (C.head_sha,))

        new_c_head = C.head_sha
        C.addCommit()
        old_c_head = C.head_sha
        self.assertNotEqual(old_c_head, new_c_head)
        self.fake_github.emitEvent(C.getPullRequestEditedEvent())
        self.waitUntilSettled()
        self.assertEqual(self.history[-1].changes, '2,%s' %
                         (C.head_sha,))

        new_a_head = A.head_sha
        A.addCommit()
        old_a_head = A.head_sha
        self.assertNotEqual(old_a_head, new_a_head)
        self.fake_github.emitEvent(A.getPullRequestEditedEvent())
        self.waitUntilSettled()
        self.assertEqual(self.history[-1].changes, '2,%s 1,1 1,%s' %
                         (C.head_sha, A.head_sha,))

    def test_crd_check_unknown(self):
        "Test unknown projects in independent pipeline"
        self.init_repo("gerrit/unknown", tag='init')
        A = self.fake_github.openFakePullRequest('github/project2', 'master',
                                                 'A')
        B = self.fake_gerrit.addFakeChange(
            'gerrit/unknown', 'master', 'B')

        # A Depends-On: B
        event = A.editBody('Depends-On: %s\n' % (B.data['url'],))

        # Make sure zuul has seen an event on B.
        self.fake_gerrit.addEvent(B.getPatchsetCreatedEvent(1))
        # Note we wait until settled here as the event processing for
        # the next event may not have the updated db yet otherwise.
        self.waitUntilSettled()
        self.fake_github.emitEvent(event)
        self.waitUntilSettled()

        self.assertFalse(A.is_merged)
        self.assertEqual(len(A.comments), 1)
        self.assertEqual(B.data['status'], 'NEW')
        self.assertEqual(B.reported, 0)

    def test_crd_cycle_join(self):
        "Test an updated change creates a cycle"
        A = self.fake_gerrit.addFakeChange(
            'gerrit/project1', 'master', 'A')

        self.fake_gerrit.addEvent(A.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()
        self.assertEqual(A.reported, 1)

        # Create B->A
        B = self.fake_github.openFakePullRequest('github/project2', 'master',
                                                 'B')
        self.fake_github.emitEvent(
            B.editBody('Depends-On: %s\n' % (A.data['url'],)))
        self.waitUntilSettled()

        # Dep is there so zuul should have reported on B
        self.assertEqual(len(B.comments), 1)

        # Update A to add A->B (a cycle).
        A.data['commitMessage'] = '%s\n\nDepends-On: %s\n' % (
            A.subject, B.url)
        self.fake_gerrit.addEvent(A.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()

        # Dependency cycle injected so zuul should have reported again on A
        self.assertEqual(A.reported, 2)

        # Now if we update B to remove the depends-on, everything
        # should be okay.  B; A->B

        B.addCommit()
        B.editBody('')
        self.fake_gerrit.addEvent(A.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()

        # Cycle was removed so now zuul should have reported again on A
        self.assertEqual(A.reported, 3)

        self.fake_github.emitEvent(B.getPullRequestEditedEvent())
        self.waitUntilSettled()
        self.assertEqual(len(B.comments), 2)
